import uuid
from importlib.resources import files

import pytest

from a3m.fpr.registry import JSONBackend
from a3m.fpr.registry import Registry
from a3m.fpr.registry import RulePurpose
from a3m.main.models import File


def create_file_with_version_id(id: str) -> File:
    f = File.objects.create(uuid=uuid.uuid4())
    f.fileformatversion_set.create(format_version_id=id)

    return f


@pytest.fixture(scope="module")
def registry():
    backend = JSONBackend(
        files("a3m.fpr.migrations").joinpath("initial-data.json").read_bytes()
    )
    return Registry(backend)


def test_registry_get_file_rules(db, registry):
    """Confirm that it accepts uuid.UUID, str and File."""
    file_obj = create_file_with_version_id(
        "082f3282-8331-4da4-b452-632b17e90d66"
    )  # fmt/3
    assert len(registry.get_file_rules(file_obj.uuid, RulePurpose.THUMBNAIL)) == 1
    assert len(registry.get_file_rules(str(file_obj.uuid), RulePurpose.THUMBNAIL)) == 1
    assert len(registry.get_file_rules(file_obj, RulePurpose.THUMBNAIL)) == 1


def test_registry_integrity(registry):
    """Validates the integrity of the registry.

    It depends on implementation details of the JSONBackend.
    """
    backend: JSONBackend = registry.backend

    # Verify that all replaced rules are marked as disabled.
    for rule in backend.rules.values():
        is_replaced = rule.id in backend.replaced_rules
        is_enabled = rule.enabled
        assert not (
            is_enabled and is_replaced
        ), f"Rule {rule.id} is enabled but has been replaced by Rule {backend.replaced_rules[rule.id]}."

    # Verify that all replaced versions are marked as disabled.
    for version in backend.versions.values():
        is_replaced = version.id in backend.replaced_versions
        is_enabled = version.enabled
        assert not (
            is_enabled and is_replaced
        ), f"FormatVersion {version.id} is enabled but has been replaced by FormatVersion {backend.replaced_versions[version.id]}."

    # Verify that all replaced commands are marked as disabled.
    for command in backend.commands.values():
        is_replaced = command.id in backend.replaced_versions
        is_enabled = command.enabled
        assert not (
            is_enabled and is_replaced
        ), f"Command {command.id} is enabled but has been replaced by Command {backend.replaced_versions[command.id]}."

    # Verify that rules in service depend on commands and versions in service.
    for rule in backend.rules.values():
        is_replaced = rule.id in backend.replaced_rules
        is_enabled = rule.enabled
        if is_enabled and not is_replaced:
            assert (
                rule.command.enabled
                and rule.command.id not in backend.replaced_commands
            ), f"Rule in service {rule.id} is using a Command not in service: {rule.command} ({rule.command.tool.description})."
            assert (
                rule.format.enabled and rule.format.id not in backend.replaced_versions
            ), f"Rule in service {rule.id} is using a FormatVersion not in service: {rule.format} ({rule.format.description})."



from . import request_response_pb2
from . import request_response_pb2_grpc
from . import service_pb2
from . import service_pb2_grpc


__all__ = [
    "request_response_pb2_grpc",
    "request_response_pb2",
    "service_pb2_grpc",
    "service_pb2",
]



import datetime
import logging
import sys

import click
from django.conf import settings
from google.protobuf.descriptor import FieldDescriptor
from rich.console import Console
from rich.panel import Panel
from rich.table import Table

from a3m.api.transferservice import v1beta1 as transfer_service_api
from a3m.cli.client.wrapper import ClientWrapper
from a3m.cli.common import configure_xml_catalog_files
from a3m.cli.common import init_django
from a3m.cli.common import suppress_warnings
from a3m.server.processing import DEFAULT_PROCESSING_CONFIG
from a3m.server.rpc.client import Client


@click.command()
@click.argument("uri")
@click.option("--name", help="Name of the package.", metavar="NAME")
@click.option(
    "--address",
    help='a3m server address (form "host:port"), e.g.: "172.26.30.2:12345".',
    metavar="ADDRESS",
)
@click.option(
    "--wait-for-ready", is_flag=True, help="Block until server becomes available."
)
@click.option(
    "--processing-config",
    "-p",
    multiple=True,
    metavar="CONFIG_PAIR",
    help='Processing configuration pair (form "name:value"), e.g.: "normalize=no".',
)
@click.option("--no-input", is_flag=True, help="Disable interactive mode.")
@click.pass_context
def main(ctx, uri, name, address, processing_config, wait_for_ready, no_input):
    """a3m - Lightweight Archivematica.

    Creates an Archival Information Package (AIP) from the contents in URI.

    URI specifies the location of the transfer. Schemes supported are file,
    http and https, e.g. "file:///mnt/disk1/pictures".

    a3m launches its own embedded instance of the engine unless `--address` is
    used to refer to a remote instance. Use `--wait-for-ready` if you want the
    client to block until the server becomes available. If you are running this
    tool in an automated fashion, use `--no-input` to avoid prompts.
    """
    init_django()
    suppress_warnings()
    configure_xml_catalog_files()

    # Disable logging.
    if not settings.DEBUG:
        logging.disable(sys.maxsize)

    # A3M-TODO: stop forcing users to provide a transfer name.
    if name is None:
        if no_input:
            name = f"transfer.{datetime.datetime.now().timestamp()}"
        else:
            name = click.prompt("Enter transfer name")

    processing_config = _prepare_config(processing_config)

    with ClientWrapper(address, wait_for_ready) as cw:
        resp = cw.client.submit(uri, name, processing_config)
        click.secho(f"AIP {resp.id} is being generated...")

        resp = cw.client.wait_until_complete(resp.id)

        if (status := resp.status) in (
            transfer_service_api.request_response_pb2.PACKAGE_STATUS_FAILED,
            transfer_service_api.request_response_pb2.PACKAGE_STATUS_REJECTED,
        ):
            click.secho(
                f"Error processing package ({transfer_service_api.request_response_pb2.PackageStatus.Name(status)})!",
                fg="red",
            )
            _print_failed_jobs(cw.client, resp.jobs)
            ctx.exit(1)

        click.secho("Processing completed successfully!", fg="green")


def _to_int(value: str) -> int | None:
    try:
        return int(value)
    except ValueError:
        return None


def _prepare_config(user_pairs):
    """Consolidate ``ProcessingConfig`` defaults and user-provided.

    ``user_pairs`` is a list of strings following the format ``key=value``.
    Those matching processing configuration attributes will be evaluated as
    indicated by the field type, e.g.:

    * ``normalize=yes`` (boolean), or
    * ``aip_compression_level=1`` (integer), or
    * ``aip_compression_algorithm=1`` (enum)
        (it is not possible at the moment to use the enum names)

    A comprehensive list can be found in the definition of the
    ``ProcessingConfig`` message in the proto file.
    """
    config = transfer_service_api.request_response_pb2.ProcessingConfig()
    config.CopyFrom(DEFAULT_PROCESSING_CONFIG)
    for item in user_pairs:
        head, sep, tail = item.partition("=")
        if head and sep:
            try:
                field = transfer_service_api.request_response_pb2.ProcessingConfig.DESCRIPTOR.fields_by_name[
                    head
                ]
            except KeyError:
                continue
            if field.type == FieldDescriptor.TYPE_BOOL:
                enabled = tail.lower() in ("yes", "true", "1", "on")
                setattr(config, head, enabled)
            elif field.type == FieldDescriptor.TYPE_INT32:
                if (value := _to_int(tail)) is not None:
                    setattr(config, head, value)
            elif field.type == FieldDescriptor.TYPE_ENUM:
                if (value := _to_int(tail)) is not None:
                    setattr(config, head, value)
            else:
                raise NotImplementedError(
                    f"{head} has an unsupported type ({field.type})"
                )
    return config


def _print_failed_jobs(client: Client, jobs):
    """Prints failed jobs and associated tasks for a failed package."""
    if not jobs:
        return
    console = Console()
    console.rule("Failed jobs")
    item: transfer_service_api.request_response_pb2.Job
    for item in jobs:
        if item.status != item.STATUS_FAILED:
            continue
        table = Table(expand=True, show_header=True, header_style="bold magenta")
        table.add_column("Job")
        table.add_column("Identifier")
        table.add_column("Link")
        table.add_row(
            f"[bold]{item.name}[/]\n[dim](part of {item.group})[/]",
            item.id,
            item.link_id,
        )
        console.print(table)
        try:
            resp = client.list_tasks(item.id)
        except Exception:
            console.print("Tasks could not be loaded.")
            continue
        task: transfer_service_api.request_response_pb2.Task
        for task in resp.tasks:
            content = f"""[bold]Task {task.id}[/]

Module [bold]{task.execution}[/] (with arguments: [dim]{task.arguments}[/])
--- stdout
{task.stdout}
--- stderr
[red]{task.stderr}[/]
"""
            console.print(Panel(content))


if __name__ == "__main__":
    main()



import logging
import os
import platform
import signal

import grpc
from django.conf import settings

from a3m import __version__
from a3m.cli.common import configure_xml_catalog_files
from a3m.cli.common import init_django
from a3m.cli.common import suppress_warnings


logger = logging.getLogger(__name__)


def main():
    init_django()
    suppress_warnings()
    configure_xml_catalog_files()

    from a3m.server.runner import create_server

    logger.info(
        f"Starting a3m... (version={__version__} pid={os.getpid()} "
        f"uid={os.getuid()} python={platform.python_version()} "
        f"listen={settings.RPC_BIND_ADDRESS})"
    )

    # A3M-TODO: make this configurable, e.g. local tcp, local uds, tls certs...
    # (see https://grpc.github.io/grpc/python/grpc.html#create-server-credentials for more)
    server_credentials = grpc.local_server_credentials(
        grpc.LocalConnectionType.LOCAL_TCP
    )

    server = create_server(
        settings.RPC_BIND_ADDRESS,
        server_credentials,
        settings.CONCURRENT_PACKAGES,
        settings.BATCH_SIZE,
        settings.WORKER_THREADS,
        settings.RPC_THREADS,
        settings.DEBUG,
    )
    server.start()

    def signal_handler(signo, frame):
        logger.info("Received termination signal (%s)", signal.Signals(signo).name)
        server.stop()

    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)

    server.wait_for_termination()

    logger.info("a3m shutdown complete.")


if __name__ == "__main__":
    main()



"""
A base class for other Job types to inherit from.
"""
import abc
import logging
import uuid

from django.utils import timezone

from a3m.main import models
from a3m.server.db import auto_close_old_connections


logger = logging.getLogger(__name__)


class Job(metaclass=abc.ABCMeta):
    """
    A single job, corresponding to a workflow link, and the `Job` model in the
    database.

    Subclasses must implement a `run` method; it will be called in a thread via
    `executor.submit`, and should return the next job to be processed.
    """

    # Mirror job model statuses, so that we can mostly avoid referencing
    # the job model
    STATUSES = models.Job.STATUS
    STATUS_UNKNOWN = models.Job.STATUS_UNKNOWN
    STATUS_COMPLETED_SUCCESSFULLY = models.Job.STATUS_COMPLETED_SUCCESSFULLY
    STATUS_EXECUTING_COMMANDS = models.Job.STATUS_EXECUTING_COMMANDS
    STATUS_FAILED = models.Job.STATUS_FAILED

    def __init__(self, job_chain, link, package):
        self.uuid = uuid.uuid4()
        self.job_chain = job_chain
        self.package = package
        self.link = link
        self.created_at = timezone.now()
        self.group = link.get_label("group", "en")
        self.description = link.get_label("description", "en")

        # always zero for non task jobs
        self.exit_code = 0

    @classmethod
    @auto_close_old_connections()
    def cleanup_old_db_entries(cls):
        """Update the status of any in progress jobs.

        This command is run on startup.
        TODO: we could try to recover, instead of just failing.
        """
        models.Job.objects.filter(currentstep=cls.STATUS_EXECUTING_COMMANDS).update(
            currentstep=cls.STATUS_FAILED
        )

    @abc.abstractmethod
    def run(self):
        """
        Run the actual job.

        This method is executed via ThreadPoolExecutor and returns the _next_ job
        to process.
        """

    @auto_close_old_connections()
    def save_to_db(self):
        return models.Job.objects.create(
            jobuuid=self.uuid,
            jobtype=self.description,
            directory=self.package.current_path_for_db,
            sipuuid=self.package.subid,
            currentstep=self.STATUS_EXECUTING_COMMANDS,
            unittype=self.package.unit_type,
            microservicegroup=self.group,
            createdtime=self.created_at,
            createdtimedec=float(self.created_at.strftime("0.%f")),
            microservicechainlink=self.link.id,
        )

    @auto_close_old_connections()
    def mark_complete(self):
        logger.debug(
            "%s %s done with exit code %s",
            self.__class__.__name__,
            self.uuid,
            self.exit_code,
        )
        return models.Job.objects.filter(jobuuid=self.uuid).update(
            currentstep=self.STATUS_COMPLETED_SUCCESSFULLY
        )



import logging
import os
import sys
import uuid

from django.db import transaction

from .has_packages import already_extracted
from a3m.archivematicaFunctions import format_subdir_path
from a3m.archivematicaFunctions import get_dir_uuids
from a3m.databaseFunctions import fileWasRemoved
from a3m.executeOrRunSubProcess import executeOrRun
from a3m.fileOperations import addFileToTransfer
from a3m.fileOperations import updateSizeAndChecksum
from a3m.fpr.registry import CommandScriptType
from a3m.fpr.registry import FPR
from a3m.fpr.registry import RulePurpose
from a3m.main.models import Directory
from a3m.main.models import File
from a3m.main.models import Transfer

logger = logging.getLogger(__name__)

TRANSFER_DIRECTORY = "%transferDirectory%"


def temporary_directory(file_path, date, file_path_cache):
    try:
        return file_path_cache[file_path], file_path_cache
    except KeyError:
        path = file_path + "-" + date
        file_path_cache[file_path] = path
        return path, file_path_cache


def tree(root):
    for dirpath, __, files in os.walk(root):
        for file in files:
            yield os.path.join(dirpath, file)


def assign_uuid(
    job,
    filename,
    extracted_file_original_location,
    package_uuid,
    transfer_uuid,
    date,
    task_uuid,
    sip_directory,
    package_filename,
):
    """Assign a uuid to each file in the extracted package."""
    file_uuid = str(uuid.uuid4())
    # Correct the information in the path strings sent to this function. First
    # remove the SIP directory from the string. Second, make sure that file
    # paths have not been modified for processing purpose, i.e. in
    # Archivematica current terminology, filename change.
    relative_path = filename.replace(sip_directory, TRANSFER_DIRECTORY, 1)
    relative_package_path = package_filename.replace(
        sip_directory, TRANSFER_DIRECTORY, 1
    )
    package_detail = f"{relative_package_path} ({package_uuid})"
    event_detail = "Unpacked from: " + package_detail
    addFileToTransfer(
        relative_path,
        file_uuid,
        transfer_uuid,
        task_uuid,
        date,
        sourceType="unpacking",
        eventDetail=event_detail,
        originalLocation=extracted_file_original_location,
    )
    updateSizeAndChecksum(file_uuid, filename, date, str(uuid.uuid4()))
    job.pyprint("Assigning new file UUID:", file_uuid, "to file", filename)


def _get_subdir_paths(job, root_path, path_prefix_to_repl, original_location):
    """Return a generator of subdirectory paths in ``root_path`` with
    the ancestor path ``path_prefix_to_repl`` replaced by a placeholder
    string. ``original_location`` should be the zip container that the content
    was extracted from as it was transferred. Between then and now the path
    will have been changed to remove characters that are difficult for tools
    to handle. PREMIS describes the original location as how it was received
    not what it was changed to, so we make these changes below to reflect the
    original transfer material.
    """
    # Make sure that the original location is suffixed with a trailing slash
    # as it is a directory.
    original_location = os.path.join(original_location, "")
    # Transform the processing directory string to reflect where it was when
    # transferred, that is, in the ``%transferDirectory%``.
    location_to_replace = format_subdir_path(root_path, path_prefix_to_repl)

    # Return a generator here that contains information about the current path
    # and the original path for the PREMIS information in the METS file.
    for dir_path, __, ___ in os.walk(root_path):
        formatted_path = format_subdir_path(dir_path, path_prefix_to_repl)
        for dir_uuid in get_dir_uuids([formatted_path], logger, printfn=job.pyprint):
            dir_uuid["originalLocation"] = formatted_path.replace(
                location_to_replace, original_location
            )
            yield dir_uuid


def delete_and_record_package_file(job, file_path, file_uuid, current_location):
    os.remove(file_path)
    job.pyprint("Package removed: " + file_path)
    event_detail_note = "removed from: " + current_location
    fileWasRemoved(file_uuid, eventDetail=event_detail_note)


def main(job, transfer_uuid, sip_directory, date, task_uuid, delete=False):
    file_path_cache = {}
    files = File.objects.filter(transfer=transfer_uuid, removedtime__isnull=True)
    if not files:
        job.pyprint("No files found for transfer: ", transfer_uuid)

    transfer_mdl = Transfer.objects.get(uuid=transfer_uuid)

    # We track whether or not anything was extracted because that controls what
    # the next microservice chain link will be.
    # If something was extracted, then a new identification step has to be
    # kicked off on those files; otherwise, we can go ahead with the transfer.
    extracted = False

    for file_obj in files:
        if not file_obj.currentlocation:
            job.pyprint(
                "Not extracting contents - currentlocation is empty", file=sys.stderr
            )
            continue
        location = os.path.basename(file_obj.currentlocation)

        format_version = file_obj.get_format_version()
        if format_version is None:
            job.pyprint(
                f"Not extracting contents from {location} - file format not identified.",
                file=sys.stderr,
            )
            continue

        rules = FPR.get_file_rules(file_obj, purpose=RulePurpose.EXTRACT)
        if not rules:
            job.pyprint(
                f"Not extracting contents from {location} - no rule found to extract.",
                file=sys.stderr,
            )
            continue

        # Check if file has already been extracted
        if already_extracted(file_obj):
            job.pyprint(
                f"Not extracting contents from {location} - extraction already happened.",
                file=sys.stderr,
            )
            continue

        file_to_be_extracted_path = file_obj.currentlocation.replace(
            TRANSFER_DIRECTORY, sip_directory
        )
        extraction_target, file_path_cache = temporary_directory(
            file_to_be_extracted_path, date, file_path_cache
        )

        rule = rules[0]
        command = rule.command

        # Create the extract packages command.
        if command.script_type in (
            CommandScriptType.COMMAND,
            CommandScriptType.BASH_SCRIPT,
        ):
            args = []
            command_to_execute = command.command.replace(
                "%inputFile%", file_to_be_extracted_path
            )
            command_to_execute = command_to_execute.replace(
                "%outputDirectory%", extraction_target
            )
        else:
            command_to_execute = command.command
            args = [file_to_be_extracted_path, extraction_target]

        # Make the command clear to users when inspecting stdin/stdout.
        logger.debug("Command to execute is: %s", command_to_execute)
        exitstatus, stdout, stderr = executeOrRun(
            command.script_type, command_to_execute, arguments=args, capture_output=True
        )
        job.write_output(stdout)
        job.write_error(stderr)

        if not exitstatus == 0:
            # Dang, looks like the extraction failed
            job.pyprint("Command", command.description, "failed!", file=sys.stderr)
            continue

        extracted = True
        job.pyprint(
            "Extracted contents from", os.path.basename(file_to_be_extracted_path)
        )

        # Assign UUIDs and insert them into the database, so the newly
        # extracted files are properly tracked by Archivematica
        for extracted_file in tree(extraction_target):
            extracted_file_original_location = extracted_file.replace(
                extraction_target, file_obj.originallocation, 1
            )
            assign_uuid(
                job,
                extracted_file,
                extracted_file_original_location,
                file_obj.uuid,
                transfer_uuid,
                date,
                task_uuid,
                sip_directory,
                file_to_be_extracted_path,
            )

        if transfer_mdl.diruuids:
            create_extracted_dir_uuids(
                job, transfer_mdl, extraction_target, sip_directory, file_obj
            )

        # We may want to remove the original package file after extracting
        # its contents
        if delete:
            delete_and_record_package_file(
                job, file_to_be_extracted_path, file_obj.uuid, file_obj.currentlocation
            )

    return 0 if extracted else 255


def create_extracted_dir_uuids(
    job, transfer_mdl, extraction_target, sip_directory, file_obj
):
    """Assign UUIDs to directories via ``Directory`` objects in the database."""
    Directory.create_many(
        dir_paths_uuids=_get_subdir_paths(
            job=job,
            root_path=extraction_target,
            path_prefix_to_repl=sip_directory,
            original_location=file_obj.originallocation,
        ),
        unit_mdl=transfer_mdl,
    )


def call(jobs):
    with transaction.atomic():
        for job in jobs:
            with job.JobContext(logger=logger):
                transfer_uuid = job.args[1]
                sip_directory = job.args[2]
                date = job.args[3]
                task_uuid = job.args[4]
                # Whether or not to remove the package file post-extraction
                # This is set by the user during the transfer, and defaults to false.
                delete = False
                if job.args[5] == "True":
                    delete = True
                job.pyprint(f"Deleting?: {delete}", file=sys.stderr)
                job.set_status(
                    main(
                        job,
                        transfer_uuid,
                        sip_directory,
                        date,
                        task_uuid,
                        delete=delete,
                    )
                )



import sys
from importlib.metadata import version


def shim_pkg_resources():
    """Injects a pkg_resources fake needed by bagit-python in Python 3.12.

    The underlying error is only reproducible if setuptools is not installed.
    """

    class Distribution:
        def __init__(self, name):
            self.version = version(name)

    class FakeDistributionNotFound(Exception):
        pass

    def fake_get_distribution(_, name):
        return Distribution(name)

    class PkgResources:
        DistributionNotFound = FakeDistributionNotFound
        get_distribution = fake_get_distribution

    sys.modules["pkg_resources"] = PkgResources()  # type: ignore


shim_pkg_resources()



"""Package management."""
import ast
import collections
import functools
import logging
import os
from dataclasses import dataclass
from dataclasses import field
from enum import auto
from enum import Enum
from uuid import uuid4

from django.conf import settings
from google.protobuf import timestamp_pb2

from a3m.api.transferservice import v1beta1 as transfer_service_api
from a3m.archivematicaFunctions import strToUnicode
from a3m.main import models
from a3m.server.db import auto_close_old_connections
from a3m.server.jobs import JobChain
from a3m.server.processing import DEFAULT_PROCESSING_CONFIG

logger = logging.getLogger(__name__)


def _get_setting(name):
    """Retrieve a Django setting decoded as a unicode string."""
    return strToUnicode(getattr(settings, name))


BASE_REPLACEMENTS = {
    r"%tmpDirectory%": os.path.join(_get_setting("SHARED_DIRECTORY"), "tmp", ""),
    r"%processingDirectory%": _get_setting("PROCESSING_DIRECTORY"),
    r"%rejectedDirectory%": _get_setting("REJECTED_DIRECTORY"),
}


def get_file_replacement_mapping(file_obj, unit_directory):
    mapping = BASE_REPLACEMENTS.copy()
    dirname = os.path.dirname(file_obj.currentlocation)
    name, ext = os.path.splitext(file_obj.currentlocation)
    name = os.path.basename(name)

    absolute_path = file_obj.currentlocation.replace(r"%SIPDirectory%", unit_directory)
    absolute_path = absolute_path.replace(r"%transferDirectory%", unit_directory)

    mapping.update(
        {
            r"%fileUUID%": file_obj.pk,
            r"%originalLocation%": file_obj.originallocation,
            r"%currentLocation%": file_obj.currentlocation,
            r"%fileGrpUse%": file_obj.filegrpuse,
            r"%fileDirectory%": dirname,
            r"%fileName%": name,
            r"%fileExtension%": ext[1:],
            r"%fileExtensionWithDot%": ext,
            r"%relativeLocation%": absolute_path,
            # TODO: standardize duplicates
            r"%inputFile%": absolute_path,
            r"%fileFullName%": absolute_path,
        }
    )

    return mapping


class Stage(Enum):
    """Package stages."""

    TRANSFER = auto()
    INGEST = auto()


class Package:
    """Package is the processing unit in a3m.

    It wraps a SIP and borrows its SIP. But it also knows about its transfer
    stage. Some methods return different values depending the stage the
    package is in.
    """

    def __init__(self, name, url, config, transfer, sip):
        self.name = name
        self.url = url
        self.config = self._prepare_config(config)
        self.transfer = transfer
        self.sip = sip
        self.stage = Stage.TRANSFER
        self.aip_filename = None
        self._current_path = self.transfer.currentlocation

    def __repr__(self):
        return "{class_name}({uuid})".format(
            class_name=self.__class__.__name__, uuid=self.uuid
        )

    def _prepare_config(self, provided=None):
        """Combine user-provided processing config with the defaults."""
        config = transfer_service_api.request_response_pb2.ProcessingConfig()
        config.CopyFrom(DEFAULT_PROCESSING_CONFIG)
        if provided is not None:
            for config_field in config.DESCRIPTOR.fields:
                field_name = config_field.name
                setattr(config, field_name, getattr(provided, field_name))

        return config

    @classmethod
    @auto_close_old_connections()
    def create_package(cls, package_queue, executor, workflow, name, url, config):
        """Launch transfer and return its object immediately."""
        if not name:
            raise ValueError("No transfer name provided.")
        if not url:
            raise ValueError("No url provided.")

        transfer_id = str(uuid4())
        transfer_dir = os.path.join(
            _get_setting("PROCESSING_DIRECTORY"), "transfer", transfer_id, ""
        )
        transfer = models.Transfer.objects.create(
            uuid=transfer_id, currentlocation=transfer_dir
        )
        logger.debug("Transfer object created: %s", transfer.pk)

        sip_id = str(uuid4())
        sip_dir = os.path.join(
            _get_setting("PROCESSING_DIRECTORY"), "ingest", sip_id, ""
        )
        sip = models.SIP.objects.create(uuid=sip_id, currentpath=sip_dir)
        sip.transfer_id = transfer_id
        logger.debug("SIP object created: %s", sip.pk)

        package = cls(name, url, config, transfer, sip)

        params = (package, package_queue, workflow)
        future = executor.submit(Package.trigger_workflow, *params)
        future.add_done_callback(
            functools.partial(Package.trigger_workflow_done_callback, package.uuid)
        )

        return package

    @staticmethod
    def trigger_workflow(package, package_queue, workflow):
        logger.debug("Package %s: starting workflow processing", package.uuid)

        initiator_link = workflow.get_initiator()
        if initiator_link is None:
            raise ValueError("Workflow initiator not found")

        job_chain = JobChain(package, workflow, initiator_link)

        package_queue.schedule_job(next(job_chain))

    @staticmethod
    def trigger_workflow_done_callback(package_id, future):
        try:
            future.result()
        except Exception as err:
            logger.warning("Exception detected: %s", err, exc_info=True)
        else:
            logger.info("Package processing started (%s)", package_id)

    @property
    def uuid(self):
        return self.sip.pk

    @property
    def subid(self):
        if self.stage is Stage.INGEST:
            return self.sip.pk
        else:
            return self.transfer.pk

    @property
    def current_path(self):
        return self._current_path

    @current_path.setter
    def current_path(self, value):
        """The real (no shared dir vars) path to the package."""
        self._current_path = value.replace(
            r"%sharedPath%", _get_setting("SHARED_DIRECTORY")
        )

    @property
    def current_path_for_db(self):
        """The path to the package, as stored in the database."""
        return self.current_path.replace(
            _get_setting("SHARED_DIRECTORY"), r"%sharedPath%", 1
        )

    @property
    def context(self):
        """Returns a `PackageContext` for this package."""
        # This needs to be reloaded from the db every time, because new values
        # could have been added by a client script.
        # TODO: pass context changes back from client
        return PackageContext.load_from_db(self.subid)

    @property
    def unit_type(self):
        if self.stage is Stage.INGEST:
            return "unitSIP"
        else:
            return "unitTransfer"

    @property
    def unit_variable_type(self):
        if self.stage is Stage.INGEST:
            return "SIP"
        else:
            return "Transfer"

    @property
    def replacement_path_string(self):
        if self.stage is Stage.INGEST:
            return r"%SIPDirectory%"
        else:
            return r"%transferDirectory%"

    @property  # type: ignore
    @auto_close_old_connections()
    def base_queryset(self):
        if self.stage is Stage.INGEST:
            return models.File.objects.filter(sip_id=self.sip.pk)
        else:
            return models.File.objects.filter(transfer_id=self.transfer.pk)

    def start_ingest(self):
        """Signal this package so it becomes a SIP."""
        self.stage = Stage.INGEST

    @auto_close_old_connections()
    def reload(self):
        if self.stage is Stage.INGEST:
            sip = models.SIP.objects.get(uuid=self.sip.pk)
            self.current_path = sip.currentpath
            self.aip_filename = sip.aip_filename or ""
        else:
            transfer = models.Transfer.objects.get(uuid=self.transfer.pk)
            self.current_path = transfer.currentlocation

    def get_replacement_mapping(self):
        mapping = BASE_REPLACEMENTS.copy()
        mapping.update(
            {
                r"%SIPUUID%": str(self.sip.pk),
                r"%TransferUUID%": str(self.transfer.pk),
                r"%SIPName%": self.name,
                r"%SIPLogsDirectory%": os.path.join(self.current_path, "logs", ""),
                r"%SIPObjectsDirectory%": os.path.join(
                    self.current_path, "objects", ""
                ),
                r"%SIPDirectory%": self.current_path,
                r"%SIPDirectoryBasename%": os.path.basename(
                    os.path.abspath(self.current_path)
                ),
                r"%relativeLocation%": self.current_path_for_db,
            }
        )

        mapping.update(
            {
                rf"%config:{config_attr.name}%": str(
                    getattr(self.config, config_attr.name)
                )
                for config_attr in transfer_service_api.request_response_pb2.ProcessingConfig.DESCRIPTOR.fields
            }
        )

        if self.stage is Stage.INGEST:
            mapping.update(
                {
                    r"%unitType%": self.unit_variable_type,
                    r"%AIPFilename%": self.aip_filename,
                }
            )
        else:
            mapping.update(
                {
                    self.replacement_path_string: self.current_path,
                    r"%unitType%": self.unit_variable_type,
                    r"%URL%": self.url,
                }
            )

        return mapping

    def files(self, filter_subdir=None):
        """Generator that yields all files associated with the package or that
        should be associated with a package.
        """
        with auto_close_old_connections():
            queryset = self.base_queryset

            if filter_subdir:
                filter_path = "".join([self.replacement_path_string, filter_subdir])
                queryset = queryset.filter(currentlocation__startswith=filter_path)

            start_path = self.current_path
            if filter_subdir:
                start_path = start_path + filter_subdir

            files_returned_already = set()
            if queryset.exists():
                for file_obj in queryset.iterator():
                    file_obj_mapped = get_file_replacement_mapping(
                        file_obj, self.current_path
                    )
                    if not os.path.exists(file_obj_mapped.get("%inputFile%")):
                        continue
                    files_returned_already.add(file_obj_mapped.get("%inputFile%"))
                    yield file_obj_mapped

            for basedir, subdirs, files in os.walk(start_path):
                for file_name in files:
                    file_path = os.path.join(basedir, file_name)
                    if file_path not in files_returned_already:
                        yield {
                            r"%relativeLocation%": file_path,
                            r"%fileUUID%": "None",
                            r"%fileGrpUse%": "",
                        }


class PackageContext:
    """Package context tracks choices made previously while processing"""

    def __init__(self, *items):
        self._data = collections.OrderedDict()
        for key, value in items:
            self._data[key] = value

    def __repr__(self):
        return f"PackageContext({dict(list(self._data.items()))!r})"

    def __iter__(self):
        yield from self._data.items()

    def __len__(self):
        return len(self._data)

    def __contains__(self, key):
        return key in self._data

    def __getitem__(self, key):
        return self._data[key]

    def __setitem__(self, key, value):
        self._data[key] = value

    def __delitem__(self, key):
        del self._data[key]

    @classmethod
    @auto_close_old_connections()
    def load_from_db(cls, uuid):
        """
        Loads a context from the UnitVariable table.
        """
        context = cls()

        # TODO: we shouldn't need one UnitVariable per chain, with all the same values
        unit_vars_queryset = models.UnitVariable.objects.filter(
            unituuid=uuid, variable="replacementDict"
        )
        # Distinct helps here, at least
        unit_vars_queryset = unit_vars_queryset.values_list("variablevalue").distinct()
        for unit_var_value in unit_vars_queryset:
            # TODO: nope nope nope, fix eval usage
            try:
                unit_var = ast.literal_eval(unit_var_value[0])
            except (ValueError, SyntaxError):
                logger.exception(
                    "Failed to eval unit variable value %s", unit_var_value[0]
                )
            else:
                context.update(unit_var)

        return context

    def copy(self):
        clone = PackageContext()
        clone._data = self._data.copy()

        return clone

    def update(self, mapping):
        for key, value in mapping.items():
            self._data[key] = value


class PackageNotFoundError(Exception):
    pass


@dataclass
class PackageStatus:
    status: int | None = None
    job: str | None = None
    jobs: list = field(default_factory=list)


@auto_close_old_connections()
def get_package_status(package_queue, package_id: str) -> PackageStatus:
    try:
        sip = models.SIP.objects.get(pk=package_id)
    except models.SIP.DoesNotExist:
        raise PackageNotFoundError

    def get_latest_job(unit_id):
        return (
            models.Job.objects.filter(sipuuid=unit_id)
            .order_by("-createdtime", "-createdtimedec")
            .first()
        )

    package = package_queue.active_packages.get(package_id)
    if package:
        # Reminder: package.subid can be in Transfer or Ingest.
        job = get_latest_job(package.subid)
        return PackageStatus(
            status=transfer_service_api.request_response_pb2.PACKAGE_STATUS_PROCESSING,
            job=job.jobtype if job else None,
        )

    # A3M-TODO: persist package-workflow status!
    # It'd be much easier if a workflow instance could keep the package
    # model(s) up to date.

    # We have an inactive package, look up the status in the database.
    job = get_latest_job(package_id)

    # It must be an error during Transfer when Ingest activity not recorded.
    if not job:
        transfer_id = sip.transfer_id
        if transfer_id is None:
            raise Exception(
                "Package status cannot be determined: transfer_id is undefined"
            )
        job = get_latest_job(sip.transfer_id)
        if job is None:
            return PackageStatus(
                status=transfer_service_api.request_response_pb2.PACKAGE_STATUS_PROCESSING
            )

    if "failed" in job.microservicegroup.lower():
        status = transfer_service_api.request_response_pb2.PACKAGE_STATUS_FAILED
    elif "reject" in job.microservicegroup.lower():
        status = transfer_service_api.request_response_pb2.PACKAGE_STATUS_REJECTED
    elif job.jobtype == "a3m - Store AIP":
        status = transfer_service_api.request_response_pb2.PACKAGE_STATUS_COMPLETE
    else:
        raise Exception(
            f"Package status cannot be determined (job.currentstep={job.currentstep}, job.type={job.jobtype}, job.microservicegroup={job.microservicegroup})"
        )

    package_status = PackageStatus(status=status, job=job.microservicegroup)

    for item in (
        models.Job.objects.filter(sipuuid__in=(sip.pk, sip.transfer_id))
        .order_by("createdtime")
        .values(
            "jobuuid",
            "jobtype",
            "currentstep",
            "microservicegroup",
            "microservicechainlink",
            "currentstep",
            "createdtime",
        )
    ):
        start_time = timestamp_pb2.Timestamp()
        start_time.FromDatetime(item["createdtime"])

        package_status.jobs.append(
            transfer_service_api.request_response_pb2.Job(
                id=str(item["jobuuid"]),
                name=item["jobtype"],
                group=item["microservicegroup"],
                link_id=str(item["microservicechainlink"]),
                status=item["currentstep"],
                start_time=start_time,
            )
        )

    return package_status



"""
Task class, representing an individual command (usually run on a file or directory).
Stored in the `Task` model.

Tasks are passed to MCPClient for processing.
"""
import logging
import os
import uuid

from django.utils import timezone

from a3m.main import models
from a3m.server.db import auto_close_old_connections


logger = logging.getLogger(__name__)


class Task:
    """A task object, representing an individual command (usually run on a file or
    directory).

    Tasks are processed out of process by a `TaskBackend`, which passes them to
    MCPClient.
    """

    def __init__(
        self,
        execute,
        arguments,
        stdout_file_path,
        stderr_file_path,
        context,
        wants_output=False,
    ):
        self.uuid = uuid.uuid4()
        self.done = False
        self.execute = execute
        self.arguments = arguments
        self.stdout_file_path = stdout_file_path
        self.stderr_file_path = stderr_file_path
        self.context = context

        self.wants_output = any([wants_output, stdout_file_path, stderr_file_path])

        self.exit_code = None
        self.stdout = ""
        self.stderr = ""

        self.start_timestamp = timezone.now()
        self.finished_timestamp = None

    def __repr__(self):
        return "Task(uuid={}, execute={}, arguments={}, start_timestamp={}, done={})".format(
            self.uuid, self.execute, self.arguments, self.start_timestamp, self.done
        )

    @classmethod
    @auto_close_old_connections()
    def cleanup_old_db_entries(cls):
        """Update the status of any in progress tasks.

        This command is run on startup.
        TODO: we could try to recover, instead of just failing.
        """
        models.Task.objects.filter(exitcode=None).update(
            exitcode=-1, stderror="MCP shut down while processing."
        )

    @classmethod
    @auto_close_old_connections()
    def bulk_log(self, tasks, job):
        """Log tasks to the database, in bulk."""
        model_objects = [task.to_db_model(job) for task in tasks]
        models.Task.objects.bulk_create(model_objects)

    def to_db_model(self, job):
        """Returns an instance of the `Task` Django model."""
        job_uuid = job.uuid
        file_uuid = self.context.get(r"%fileUUID%", "")
        task_exec = job.link.config.get("execute")
        file_name = os.path.basename(
            os.path.abspath(self.context[r"%relativeLocation%"])
        )

        return models.Task(
            taskuuid=self.uuid,
            job_id=job_uuid,
            fileuuid=file_uuid,
            filename=file_name,
            execution=task_exec,
            arguments=self.arguments,
            createdtime=self.start_timestamp,
        )

    def _write_file_to_disk(self, path, contents):
        """Write the bytes in ``contents`` to ``path`` in append mode.

        The mode of ``path`` is adjusted to ensure that it's not readable by
        ``others``.
        """
        try:
            with open(path, "a") as f:
                f.write(contents)
            os.chmod(path, 0o640)
        except Exception:
            logger.exception("Unable to write to: %s", path)

    def write_output(self):
        """
        Write the stdout/stderror we got from MCP Client out to files,
        if necessary.
        """
        if self.stdout_file_path and self.stdout:
            self._write_file_to_disk(self.stdout_file_path, self.stdout)
        if self.stderr_file_path and self.stderr:
            self._write_file_to_disk(self.stderr_file_path, self.stderr)



from a3m.fpr.registry import FPR
from a3m.fpr.registry import RulePurpose
from a3m.main.models import Event
from a3m.main.models import File
from a3m.main.models import Transfer


def is_extractable(f) -> bool:
    """Determine whether the package is extractable."""
    return len(FPR.get_file_rules(f, purpose=RulePurpose.EXTRACT)) > 0


def already_extracted(f) -> bool:
    """Determine whether the package has already been extracted."""
    # Look for files in a directory that starts with the package name
    files = File.objects.filter(
        transfer=f.transfer,
        currentlocation__startswith=f.currentlocation,
        removedtime__isnull=True,
    ).exclude(uuid=f.uuid)
    # Check for unpacking events that reference the package
    if Event.objects.filter(
        file_uuid__in=files,
        event_type="unpacking",
        event_detail__contains=f.currentlocation,
    ).exists():
        return True
    return False


def main(job, sip_uuid):
    transfer = Transfer.objects.get(uuid=sip_uuid)
    for f in transfer.file_set.filter(removedtime__isnull=True).iterator():
        if is_extractable(f) and not already_extracted(f):
            job.pyprint(
                f.currentlocation,
                "is extractable and has not yet been extracted.",
            )
            return 0
    job.pyprint("No extractable files found.")
    return 1


def call(jobs):
    for job in jobs:
        with job.JobContext():
            job.set_status(main(job, job.args[1]))



# This file is part of Archivematica.
#
# Copyright 2010-2018 Artefactual Systems Inc. <http://artefactual.com>
#
# Archivematica is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Archivematica is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Archivematica.  If not, see <http://www.gnu.org/licenses/>.
"""Test settings and globals."""
from .common import *


DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": str(get_data_dir() / "db.sqlite"),
        "TEST": {"NAME": str(get_data_dir() / "dbtest.sqlite")},
    }
}



from django.db.backends.signals import connection_created


def activate_wal_mode(sender, connection, **kwargs):
    if connection.vendor == "sqlite":
        cursor = connection.cursor()
        cursor.execute("PRAGMA journal_mode=WAL")


connection_created.connect(activate_wal_mode)



import os

from a3m.executeOrRunSubProcess import executeOrRun


def call(jobs):
    for job in jobs:
        with job.JobContext():
            sip_dir = job.args[1]
            sip_name = job.args[2]

            source_dir = os.path.join(
                sip_dir, sip_name, "data", "objects", "submissionDocumentation"
            )
            submission_docs_dir = os.path.join(sip_dir, "submissionDocumentation")

            os.makedirs(submission_docs_dir, mode=0o770, exist_ok=True)

            exit_code, std_out, std_error = executeOrRun(
                "command",
                ["cp", "-R", source_dir, submission_docs_dir],
                capture_output=True,
            )

            job.write_error(std_error)
            job.write_output(std_out)
            job.set_status(exit_code)



import os
import shutil
import sys
from pathlib import Path

from django.conf import settings
from django.db import transaction

from a3m import archivematicaFunctions
from a3m.main.models import Directory
from a3m.main.models import File
from a3m.main.models import SIP


def main(job, transfer_id, sip_id):
    processing_dir = Path(settings.PROCESSING_DIRECTORY)
    transfer_dir = processing_dir / "transfer" / transfer_id
    objects_dir = transfer_dir / "objects"
    sip_dir = processing_dir / "ingest" / sip_id

    archivematicaFunctions.create_structured_directory(sip_dir)

    # Find out if any ``Directory`` models were created for the source
    # ``Transfer``. If so, this fact gets recorded in the new ``SIP`` model.
    dir_mdls = Directory.objects.filter(
        transfer_id=transfer_id,
        currentlocation__startswith="%transferDirectory%objects",
    )
    diruuids = dir_mdls.count() > 0

    # Update model
    sip = SIP.objects.get(pk=sip_id)
    sip.diruuids = diruuids
    sip.save()

    # Move objects to SIP
    for item in os.listdir(str(objects_dir)):
        src_path = os.path.join(str(objects_dir), item)
        dst_path = os.path.join(str(sip_dir), "objects", item)
        # If dst_path already exists and is a directory, shutil.move
        # will move src_path into it rather than overwriting it;
        # to avoid incorrectly-nested paths, move src_path's contents
        # into it instead.
        if os.path.exists(dst_path) and os.path.isdir(src_path):
            for subitem in os.listdir(src_path):
                shutil.move(os.path.join(src_path, subitem), dst_path)
        else:
            shutil.move(src_path, dst_path)

    # Get the ``Directory`` models representing the subdirectories in the
    # objects/ directory. For each subdirectory, confirm it's in the SIP
    # objects/ directory, and update the current location and owning SIP.
    for dir_mdl in dir_mdls:
        currentPath = dir_mdl.currentlocation
        currentSIPDirPath = currentPath.replace(
            "%transferDirectory%", str(sip_dir) + os.sep
        )
        if os.path.isdir(currentSIPDirPath):
            dir_mdl.currentlocation = currentPath.replace(
                "%transferDirectory%", "%SIPDirectory%"
            )
            dir_mdl.sip = sip
            dir_mdl.save()
        else:
            job.pyprint("Directory not found: ", currentSIPDirPath, file=sys.stderr)

    # Get the database list of files in the objects directory.
    # For each file, confirm it's in the SIP objects directory, and update the
    # current location/ owning SIP'
    files = File.objects.filter(
        transfer_id=transfer_id,
        currentlocation__startswith="%transferDirectory%objects",
        removedtime__isnull=True,
    )
    for f in files:
        currentPath = f.currentlocation
        currentSIPFilePath = currentPath.replace(
            "%transferDirectory%", str(sip_dir) + os.sep
        )
        if os.path.isfile(currentSIPFilePath):
            f.currentlocation = currentPath.replace(
                "%transferDirectory%", "%SIPDirectory%"
            )
            f.sip = sip
            f.save()
        else:
            job.pyprint("File not found: ", currentSIPFilePath, file=sys.stderr)

    archivematicaFunctions.create_directories(
        archivematicaFunctions.MANUAL_NORMALIZATION_DIRECTORIES, basepath=str(sip_dir)
    )

    # Copy the JSON metadata file, if present; this contains a
    # serialized copy of DC metadata entered in the dashboard UI
    # during the transfer.
    src = transfer_dir / "metadata" / "dc.json"
    if src.exists():
        dst = sip_dir / "metadata" / "dc.json"
        shutil.copy(str(src), str(dst))


def call(jobs):
    with transaction.atomic():
        job = jobs[0]
        with job.JobContext():
            transfer_id = job.args[1]
            sip_id = job.args[2]

            job.set_status(main(job, transfer_id, sip_id))



import logging

from google.protobuf import timestamp_pb2
from google.rpc import code_pb2

from a3m.api.transferservice import v1beta1 as transfer_service_api
from a3m.main.models import Task
from a3m.server import shared_dirs
from a3m.server.packages import get_package_status
from a3m.server.packages import Package
from a3m.server.packages import PackageNotFoundError

logger = logging.getLogger(__name__)


class TransferService(transfer_service_api.service_pb2_grpc.TransferServiceServicer):
    def __init__(self, workflow, package_queue, executor):
        self.workflow = workflow
        self.package_queue = package_queue
        self.executor = executor

    def Submit(self, request, context):
        config = request.config if request.HasField("config") else None
        try:
            package = Package.create_package(
                self.package_queue,
                self.executor,
                self.workflow,
                request.name,
                request.url,
                config,
            )
        except Exception as err:
            logger.warning("TransferService.Submit handler error: %s", err)
            context.abort(code_pb2.INTERNAL, "Unknown error")
        return transfer_service_api.request_response_pb2.SubmitResponse(
            id=str(package.uuid)
        )

    def Read(self, request, context):
        try:
            package_status = get_package_status(self.package_queue, request.id)
        except PackageNotFoundError:
            context.abort(code_pb2.NOT_FOUND, "Package not found")
        except Exception as err:
            logger.warning("TransferService.Status handler error: %s", err)
            context.abort(code_pb2.INTERNAL, "Unknown error")
        resp = transfer_service_api.request_response_pb2.ReadResponse(
            status=package_status.status
        )
        if package_status.job:
            resp.job = package_status.job
        if package_status.jobs:
            resp.jobs.extend(package_status.jobs)
        return resp

    def ListTasks(self, request, context):
        if not request.job_id:
            context.abort(code_pb2.INVALID_ARGUMENT, "job_id is mandatory")
        resp = transfer_service_api.request_response_pb2.ListTasksResponse()
        for item in Task.objects.filter(job_id=request.job_id):
            start_time = timestamp_pb2.Timestamp()
            start_time.FromDatetime(item.starttime)
            end_time = timestamp_pb2.Timestamp()
            end_time.FromDatetime(item.endtime)
            resp.tasks.append(
                transfer_service_api.request_response_pb2.Task(
                    id=item.pk,
                    file_id=item.fileuuid,
                    exit_code=item.exitcode,
                    filename=item.filename,
                    execution=item.execution,
                    arguments=item.arguments,
                    stdout=item.stdout,
                    stderr=item.stderror,
                    start_time=start_time,
                    end_time=end_time,
                )
            )
        return resp

    def Empty(self, request, context):
        # TODO: Add check: files should not be deleted if a3m is currently processing.
        resp = transfer_service_api.request_response_pb2.EmptyResponse()
        shared_dirs.empty()
        shared_dirs.create()
        return resp



# This file is part of Archivematica.
#
# Copyright 2010-2013 Artefactual Systems Inc. <http://artefactual.com>
#
# Archivematica is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Archivematica is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Archivematica.  If not, see <http://www.gnu.org/licenses/>.
"""Assign a UUID to each file in the target directory.

This client script assigns a UUID to a file by generating a new UUID and
creating several database rows (model instances), among them a ``File``
instance recording the UUID associated to a unit model, i.e., to a ``Transfer``
or ``SIP`` instance. ``Event`` instances are also created for "ingestion" and
"accession" events.

Salient parameters are the UUID of the containing unit (Transfer or SIP), the
path to the SIP directory, and the subdirectory being targeted if any.

"""
import argparse
import logging
import os
import uuid

from django.db import transaction

from a3m.archivematicaFunctions import chunk_iterable
from a3m.fileOperations import addFileToSIP
from a3m.fileOperations import addFileToTransfer
from a3m.main.models import File

logger = logging.getLogger(__name__)


TRANSFER = "Transfer"
SIP = "SIP"


def get_transfer_file_queryset(transfer_uuid):
    """Return a queryset for File objects related to the Transfer."""
    return File.objects.filter(transfer=transfer_uuid)


def assign_transfer_file_uuid(
    job,
    file_path,
    target_dir,
    date="",
    event_uuid=None,
    sip_directory="",
    transfer_uuid=None,
    sip_uuid=None,
    use="original",
    update_use=True,
    filter_subdir=None,
):
    """Walk transaction directory and write files to database.

    If files are in a re-ingested Archivematica AIP, parse the METS
    file and reuse existing information. Otherwise, create a new UUID.

    We open a database transaction for each chunk of 10 files, in an
    attempt to balance performance with reasonable transaction lengths.
    """
    if isinstance(file_path, bytes):
        file_path = file_path.decode("utf-8")

    file_path_relative_to_sip = file_path.replace(
        sip_directory, "%transferDirectory%", 1
    )
    event_type = "ingestion"
    file_uuid = str(uuid.uuid4())
    job.print_output(f"Generated UUID for file {file_uuid}")

    addFileToTransfer(
        file_path_relative_to_sip,
        file_uuid,
        transfer_uuid,
        event_uuid,
        date,
        use=use,
        sourceType=event_type,
    )


def assign_sip_file_uuid(
    job,
    file_path,
    target_dir,
    date="",
    event_uuid=None,
    sip_directory="",
    transfer_uuid=None,
    sip_uuid=None,
    use="original",
    update_use=True,
    filter_subdir=None,
):
    """Write SIP file to database with new UUID."""
    if isinstance(file_path, bytes):
        file_path = file_path.decode("utf-8")

    file_uuid = str(uuid.uuid4())
    file_path_relative_to_sip = file_path.replace(sip_directory, "%SIPDirectory%", 1)

    matching_file = File.objects.filter(
        currentlocation=file_path_relative_to_sip,
        sip=sip_uuid,
    ).first()
    if matching_file:
        job.print_error(f"File already has UUID: {matching_file.uuid}")
        if update_use:
            matching_file.filegrpuse = use
            matching_file.save()
        return

    job.print_output(f"Generated UUID for file {file_uuid}.")
    addFileToSIP(
        file_path_relative_to_sip,
        file_uuid,
        sip_uuid,
        event_uuid,
        date,
        use=use,
    )


def assign_uuids_to_files_in_dir(**kwargs):
    """Walk target directory and write files to database with new UUID.

    We open a database transaction for each chunk of 10 files, in an
    attempt to balance performance with reasonable transaction lengths.
    """
    target_dir = kwargs["target_dir"]
    transfer_uuid = kwargs["transfer_uuid"]
    for root, _, filenames in os.walk(target_dir):
        for file_chunk in chunk_iterable(filenames):
            with transaction.atomic():
                for filename in file_chunk:
                    if not filename:
                        continue
                    kwargs["file_path"] = os.path.join(root, filename)
                    if transfer_uuid:
                        assign_transfer_file_uuid(**kwargs)
                    else:
                        assign_sip_file_uuid(**kwargs)
    return 0


def call(jobs):
    parser = argparse.ArgumentParser()
    parser.add_argument("-d", "--date", action="store", dest="date", default="")
    parser.add_argument(
        "-u",
        "--eventIdentifierUUID",
        type=uuid.UUID,
        dest="event_uuid",
    )
    parser.add_argument(
        "-s", "--sipDirectory", action="store", dest="sip_directory", default=""
    )
    parser.add_argument("-S", "--sipUUID", type=uuid.UUID, dest="sip_uuid")
    parser.add_argument("-T", "--transferUUID", type=uuid.UUID, dest="transfer_uuid")
    parser.add_argument("-e", "--use", action="store", dest="use", default="original")
    parser.add_argument(
        "--filterSubdir", action="store", dest="filter_subdir", default=None
    )
    parser.add_argument(
        "--disable-update-filegrpuse",
        action="store_false",
        dest="update_use",
        default=True,
    )

    for job in jobs:
        with job.JobContext(logger=logger):
            kwargs = vars(parser.parse_args(job.args[1:]))
            kwargs["job"] = job

            TRANSFER_SIP_UUIDS = [kwargs["sip_uuid"], kwargs["transfer_uuid"]]
            if all(TRANSFER_SIP_UUIDS) or not any(TRANSFER_SIP_UUIDS):
                job.print_error("SIP exclusive-or Transfer UUID must be defined")
                job.set_status(2)
                return

            kwargs["target_dir"] = kwargs["sip_directory"]
            if kwargs["filter_subdir"]:
                kwargs["target_dir"] = os.path.join(
                    kwargs["sip_directory"], kwargs["filter_subdir"]
                )

            job.set_status(assign_uuids_to_files_in_dir(**kwargs))



# Generated by Django 4.2.6 on 2023-10-23 20:02
import uuid

import django.db.models.deletion
from django.db import migrations
from django.db import models

import a3m.main.models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Agent",
            fields=[
                (
                    "id",
                    models.AutoField(
                        db_column="pk",
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "identifiertype",
                    models.TextField(
                        db_column="agentIdentifierType",
                        null=True,
                        verbose_name="Agent Identifier Type",
                    ),
                ),
                (
                    "identifiervalue",
                    models.TextField(
                        db_column="agentIdentifierValue",
                        help_text="Used for premis:agentIdentifierValue and premis:linkingAgentIdentifierValue in the METS file.",
                        null=True,
                        verbose_name="Agent Identifier Value",
                    ),
                ),
                (
                    "name",
                    models.TextField(
                        db_column="agentName",
                        help_text="Used for premis:agentName in the METS file.",
                        null=True,
                        verbose_name="Agent Name",
                    ),
                ),
                (
                    "agenttype",
                    models.TextField(
                        db_column="agentType",
                        default="organization",
                        help_text="Used for premis:agentType in the METS file.",
                        verbose_name="Agent Type",
                    ),
                ),
            ],
            options={
                "db_table": "Agents",
            },
        ),
        migrations.CreateModel(
            name="Derivation",
            fields=[
                (
                    "id",
                    models.AutoField(
                        db_column="pk",
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
            ],
            options={
                "db_table": "Derivations",
            },
        ),
        migrations.CreateModel(
            name="Directory",
            fields=[
                (
                    "uuid",
                    models.CharField(
                        db_column="directoryUUID",
                        max_length=36,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "originallocation",
                    a3m.main.models.BlobTextField(db_column="originalLocation"),
                ),
                (
                    "currentlocation",
                    a3m.main.models.BlobTextField(
                        db_column="currentLocation", null=True
                    ),
                ),
                (
                    "enteredsystem",
                    models.DateTimeField(auto_now_add=True, db_column="enteredSystem"),
                ),
            ],
            options={
                "db_table": "Directories",
            },
        ),
        migrations.CreateModel(
            name="DublinCore",
            fields=[
                (
                    "id",
                    models.AutoField(db_column="pk", primary_key=True, serialize=False),
                ),
                (
                    "metadataappliestoidentifier",
                    models.CharField(
                        blank=True,
                        db_column="metadataAppliesToidentifier",
                        default=None,
                        max_length=36,
                        null=True,
                    ),
                ),
                ("title", models.TextField(blank=True, db_column="title")),
                ("creator", models.TextField(blank=True, db_column="creator")),
                ("subject", models.TextField(blank=True, db_column="subject")),
                ("description", models.TextField(blank=True, db_column="description")),
                ("publisher", models.TextField(blank=True, db_column="publisher")),
                ("contributor", models.TextField(blank=True, db_column="contributor")),
                (
                    "date",
                    models.TextField(
                        blank=True,
                        db_column="date",
                        help_text="Use ISO 8601 (YYYY-MM-DD or YYYY-MM-DD/YYYY-MM-DD)",
                    ),
                ),
                ("type", models.TextField(blank=True, db_column="type")),
                ("format", models.TextField(blank=True, db_column="format")),
                ("identifier", models.TextField(blank=True, db_column="identifier")),
                ("source", models.TextField(blank=True, db_column="source")),
                ("relation", models.TextField(blank=True, db_column="relation")),
                (
                    "language",
                    models.TextField(
                        blank=True, db_column="language", help_text="Use ISO 639"
                    ),
                ),
                ("coverage", models.TextField(blank=True, db_column="coverage")),
                ("rights", models.TextField(blank=True, db_column="rights")),
                (
                    "status",
                    models.CharField(
                        choices=[("ORIGINAL", "original"), ("UPDATED", "updated")],
                        db_column="status",
                        default="ORIGINAL",
                        max_length=8,
                    ),
                ),
            ],
            options={
                "db_table": "Dublincore",
            },
        ),
        migrations.CreateModel(
            name="Event",
            fields=[
                (
                    "id",
                    models.AutoField(
                        db_column="pk",
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "event_id",
                    models.UUIDField(
                        db_column="eventIdentifierUUID",
                        default=uuid.uuid4,
                        null=True,
                        unique=True,
                    ),
                ),
                ("event_type", models.TextField(blank=True, db_column="eventType")),
                (
                    "event_datetime",
                    models.DateTimeField(auto_now=True, db_column="eventDateTime"),
                ),
                ("event_detail", models.TextField(blank=True, db_column="eventDetail")),
                (
                    "event_outcome",
                    models.TextField(blank=True, db_column="eventOutcome"),
                ),
                (
                    "event_outcome_detail",
                    models.TextField(blank=True, db_column="eventOutcomeDetailNote"),
                ),
            ],
            options={
                "db_table": "Events",
            },
        ),
        migrations.CreateModel(
            name="File",
            fields=[
                (
                    "uuid",
                    models.CharField(
                        db_column="fileUUID",
                        max_length=36,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "originallocation",
                    a3m.main.models.BlobTextField(db_column="originalLocation"),
                ),
                (
                    "currentlocation",
                    a3m.main.models.BlobTextField(
                        db_column="currentLocation", null=True
                    ),
                ),
                (
                    "filegrpuse",
                    models.CharField(
                        db_column="fileGrpUse", default="Original", max_length=50
                    ),
                ),
                (
                    "filegrpuuid",
                    models.CharField(
                        blank=True, db_column="fileGrpUUID", max_length=36
                    ),
                ),
                (
                    "checksum",
                    models.CharField(blank=True, db_column="checksum", max_length=128),
                ),
                (
                    "checksumtype",
                    models.CharField(
                        blank=True, db_column="checksumType", max_length=36
                    ),
                ),
                (
                    "size",
                    models.BigIntegerField(blank=True, db_column="fileSize", null=True),
                ),
                ("label", models.TextField(blank=True)),
                (
                    "modificationtime",
                    models.DateTimeField(
                        auto_now_add=True, db_column="modificationTime", null=True
                    ),
                ),
                (
                    "enteredsystem",
                    models.DateTimeField(auto_now_add=True, db_column="enteredSystem"),
                ),
                (
                    "removedtime",
                    models.DateTimeField(
                        db_column="removedTime", default=None, null=True
                    ),
                ),
            ],
            options={
                "db_table": "Files",
            },
        ),
        migrations.CreateModel(
            name="FileFormatVersion",
            fields=[
                (
                    "id",
                    models.AutoField(
                        db_column="pk",
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("format_version_id", models.UUIDField(editable=False)),
            ],
            options={
                "db_table": "FilesIdentifiedIDs",
            },
        ),
        migrations.CreateModel(
            name="FileID",
            fields=[
                (
                    "id",
                    models.AutoField(db_column="pk", primary_key=True, serialize=False),
                ),
                ("format_name", models.TextField(blank=True, db_column="formatName")),
                (
                    "format_version",
                    models.TextField(blank=True, db_column="formatVersion"),
                ),
                (
                    "format_registry_name",
                    models.TextField(blank=True, db_column="formatRegistryName"),
                ),
                (
                    "format_registry_key",
                    models.TextField(blank=True, db_column="formatRegistryKey"),
                ),
            ],
            options={
                "db_table": "FilesIDs",
            },
        ),
        migrations.CreateModel(
            name="FPCommandOutput",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("content", models.TextField(null=True)),
                ("rule_id", models.UUIDField(editable=False)),
            ],
        ),
        migrations.CreateModel(
            name="Identifier",
            fields=[
                (
                    "id",
                    models.AutoField(
                        db_column="pk",
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("type", models.TextField(null=True, verbose_name="Identifier Type")),
                (
                    "value",
                    models.TextField(
                        help_text="Used for premis:objectIdentifierType and premis:objectIdentifierValue in the METS file.",
                        null=True,
                        verbose_name="Identifier Value",
                    ),
                ),
            ],
            options={
                "db_table": "Identifiers",
            },
        ),
        migrations.CreateModel(
            name="Job",
            fields=[
                (
                    "jobuuid",
                    models.UUIDField(
                        db_column="jobUUID",
                        default=uuid.uuid4,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "jobtype",
                    models.CharField(blank=True, db_column="jobType", max_length=250),
                ),
                ("createdtime", models.DateTimeField(db_column="createdTime")),
                (
                    "createdtimedec",
                    models.DecimalField(
                        db_column="createdTimeDec",
                        decimal_places=10,
                        default=0.0,
                        max_digits=26,
                    ),
                ),
                ("directory", models.TextField(blank=True)),
                (
                    "sipuuid",
                    models.CharField(db_column="SIPUUID", db_index=True, max_length=36),
                ),
                (
                    "unittype",
                    models.CharField(blank=True, db_column="unitType", max_length=50),
                ),
                (
                    "currentstep",
                    models.IntegerField(
                        choices=[
                            (0, "Unknown"),
                            (1, "Completed successfully"),
                            (2, "Executing command(s)"),
                            (3, "Failed"),
                        ],
                        db_column="currentStep",
                        default=0,
                    ),
                ),
                (
                    "microservicegroup",
                    models.CharField(
                        blank=True, db_column="microserviceGroup", max_length=50
                    ),
                ),
                ("hidden", models.BooleanField(default=False)),
                (
                    "microservicechainlink",
                    models.UUIDField(
                        blank=True,
                        db_column="MicroServiceChainLinksPK",
                        default=uuid.uuid4,
                        null=True,
                    ),
                ),
            ],
            options={
                "db_table": "Jobs",
            },
        ),
        migrations.CreateModel(
            name="MetadataAppliesToType",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        db_column="pk",
                        default=uuid.uuid4,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "description",
                    models.CharField(db_column="description", max_length=50),
                ),
                (
                    "replaces",
                    models.CharField(
                        blank=True, db_column="replaces", max_length=36, null=True
                    ),
                ),
                (
                    "lastmodified",
                    models.DateTimeField(auto_now=True, db_column="lastModified"),
                ),
            ],
            options={
                "db_table": "MetadataAppliesToTypes",
            },
        ),
        migrations.CreateModel(
            name="RightsStatement",
            fields=[
                (
                    "id",
                    models.AutoField(db_column="pk", primary_key=True, serialize=False),
                ),
                (
                    "metadataappliestoidentifier",
                    models.CharField(
                        blank=True,
                        db_column="metadataAppliesToidentifier",
                        max_length=36,
                    ),
                ),
                (
                    "rightsstatementidentifiertype",
                    models.TextField(
                        blank=True,
                        db_column="rightsStatementIdentifierType",
                        verbose_name="Type",
                    ),
                ),
                (
                    "rightsstatementidentifiervalue",
                    models.TextField(
                        blank=True,
                        db_column="rightsStatementIdentifierValue",
                        verbose_name="Value",
                    ),
                ),
                (
                    "rightsholder",
                    models.IntegerField(
                        db_column="fkAgent", default=0, verbose_name="Rights holder"
                    ),
                ),
                (
                    "rightsbasis",
                    models.CharField(
                        choices=[
                            ("Copyright", "Copyright"),
                            ("Statute", "Statute"),
                            ("License", "License"),
                            ("Donor", "Donor"),
                            ("Policy", "Policy"),
                            ("Other", "Other"),
                        ],
                        db_column="rightsBasis",
                        default="Copyright",
                        max_length=64,
                        verbose_name="Basis",
                    ),
                ),
                (
                    "status",
                    models.CharField(
                        choices=[("ORIGINAL", "original"), ("UPDATED", "updated")],
                        db_column="status",
                        default="ORIGINAL",
                        max_length=8,
                    ),
                ),
                (
                    "metadataappliestotype",
                    models.ForeignKey(
                        db_column="metadataAppliesToType",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="main.metadataappliestotype",
                    ),
                ),
            ],
            options={
                "verbose_name": "Rights Statement",
                "db_table": "RightsStatement",
            },
        ),
        migrations.CreateModel(
            name="RightsStatementCopyright",
            fields=[
                (
                    "id",
                    models.AutoField(
                        db_column="pk",
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "copyrightstatus",
                    models.TextField(
                        choices=[
                            ("copyrighted", "copyrighted"),
                            ("public domain", "public domain"),
                            ("unknown", "unknown"),
                        ],
                        db_column="copyrightStatus",
                        default="unknown",
                        verbose_name="Copyright status",
                    ),
                ),
                (
                    "copyrightjurisdiction",
                    models.TextField(
                        db_column="copyrightJurisdiction",
                        verbose_name="Copyright jurisdiction",
                    ),
                ),
                (
                    "copyrightstatusdeterminationdate",
                    models.TextField(
                        blank=True,
                        db_column="copyrightStatusDeterminationDate",
                        help_text="Use ISO 8061 (YYYY-MM-DD)",
                        null=True,
                        verbose_name="Copyright determination date",
                    ),
                ),
                (
                    "copyrightapplicablestartdate",
                    models.TextField(
                        blank=True,
                        db_column="copyrightApplicableStartDate",
                        help_text="Use ISO 8061 (YYYY-MM-DD)",
                        null=True,
                        verbose_name="Copyright start date",
                    ),
                ),
                (
                    "copyrightapplicableenddate",
                    models.TextField(
                        blank=True,
                        db_column="copyrightApplicableEndDate",
                        help_text="Use ISO 8061 (YYYY-MM-DD)",
                        null=True,
                        verbose_name="Copyright end date",
                    ),
                ),
                (
                    "copyrightenddateopen",
                    models.BooleanField(
                        db_column="copyrightApplicableEndDateOpen",
                        default=False,
                        help_text="Indicate end date is open",
                        verbose_name="Open End Date",
                    ),
                ),
                (
                    "rightsstatement",
                    models.ForeignKey(
                        db_column="fkRightsStatement",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="main.rightsstatement",
                    ),
                ),
            ],
            options={
                "verbose_name": "Rights: Copyright",
                "db_table": "RightsStatementCopyright",
            },
        ),
        migrations.CreateModel(
            name="RightsStatementLicense",
            fields=[
                (
                    "id",
                    models.AutoField(
                        db_column="pk",
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "licenseterms",
                    models.TextField(
                        blank=True,
                        db_column="licenseTerms",
                        null=True,
                        verbose_name="License terms",
                    ),
                ),
                (
                    "licenseapplicablestartdate",
                    models.TextField(
                        blank=True,
                        db_column="licenseApplicableStartDate",
                        help_text="Use ISO 8061 (YYYY-MM-DD)",
                        null=True,
                        verbose_name="License start date",
                    ),
                ),
                (
                    "licenseapplicableenddate",
                    models.TextField(
                        blank=True,
                        db_column="licenseApplicableEndDate",
                        help_text="Use ISO 8061 (YYYY-MM-DD)",
                        null=True,
                        verbose_name="License end date",
                    ),
                ),
                (
                    "licenseenddateopen",
                    models.BooleanField(
                        db_column="licenseApplicableEndDateOpen",
                        default=False,
                        help_text="Indicate end date is open",
                        verbose_name="Open End Date",
                    ),
                ),
                (
                    "rightsstatement",
                    models.ForeignKey(
                        db_column="fkRightsStatement",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="main.rightsstatement",
                    ),
                ),
            ],
            options={
                "verbose_name": "Rights: License",
                "db_table": "RightsStatementLicense",
            },
        ),
        migrations.CreateModel(
            name="RightsStatementOtherRightsInformation",
            fields=[
                (
                    "id",
                    models.AutoField(
                        db_column="pk",
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "otherrightsbasis",
                    models.TextField(
                        db_column="otherRightsBasis",
                        default="Other",
                        verbose_name="Other rights basis",
                    ),
                ),
                (
                    "otherrightsapplicablestartdate",
                    models.TextField(
                        blank=True,
                        db_column="otherRightsApplicableStartDate",
                        help_text="Use ISO 8061 (YYYY-MM-DD)",
                        null=True,
                        verbose_name="Other rights start date",
                    ),
                ),
                (
                    "otherrightsapplicableenddate",
                    models.TextField(
                        blank=True,
                        db_column="otherRightsApplicableEndDate",
                        help_text="Use ISO 8061 (YYYY-MM-DD)",
                        null=True,
                        verbose_name="Other rights end date",
                    ),
                ),
                (
                    "otherrightsenddateopen",
                    models.BooleanField(
                        db_column="otherRightsApplicableEndDateOpen",
                        default=False,
                        help_text="Indicate end date is open",
                        verbose_name="Open End Date",
                    ),
                ),
                (
                    "rightsstatement",
                    models.ForeignKey(
                        db_column="fkRightsStatement",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="main.rightsstatement",
                    ),
                ),
            ],
            options={
                "verbose_name": "Rights: Other",
                "db_table": "RightsStatementOtherRightsInformation",
            },
        ),
        migrations.CreateModel(
            name="RightsStatementRightsGranted",
            fields=[
                (
                    "id",
                    models.AutoField(db_column="pk", primary_key=True, serialize=False),
                ),
                ("act", models.TextField(db_column="act")),
                (
                    "startdate",
                    models.TextField(
                        blank=True,
                        db_column="startDate",
                        help_text="Use ISO 8061 (YYYY-MM-DD)",
                        null=True,
                        verbose_name="Start",
                    ),
                ),
                (
                    "enddate",
                    models.TextField(
                        blank=True,
                        db_column="endDate",
                        help_text="Use ISO 8061 (YYYY-MM-DD)",
                        null=True,
                        verbose_name="End",
                    ),
                ),
                (
                    "enddateopen",
                    models.BooleanField(
                        db_column="endDateOpen",
                        default=False,
                        help_text="Indicate end date is open",
                        verbose_name="Open End Date",
                    ),
                ),
                (
                    "rightsstatement",
                    models.ForeignKey(
                        db_column="fkRightsStatement",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="main.rightsstatement",
                    ),
                ),
            ],
            options={
                "verbose_name": "Rights: Granted",
                "db_table": "RightsStatementRightsGranted",
            },
        ),
        migrations.CreateModel(
            name="RightsStatementStatuteInformation",
            fields=[
                (
                    "id",
                    models.AutoField(db_column="pk", primary_key=True, serialize=False),
                ),
                (
                    "statutejurisdiction",
                    models.TextField(
                        db_column="statuteJurisdiction",
                        verbose_name="Statute jurisdiction",
                    ),
                ),
                (
                    "statutecitation",
                    models.TextField(
                        db_column="statuteCitation", verbose_name="Statute citation"
                    ),
                ),
                (
                    "statutedeterminationdate",
                    models.TextField(
                        blank=True,
                        db_column="statuteInformationDeterminationDate",
                        help_text="Use ISO 8061 (YYYY-MM-DD)",
                        null=True,
                        verbose_name="Statute determination date",
                    ),
                ),
                (
                    "statuteapplicablestartdate",
                    models.TextField(
                        blank=True,
                        db_column="statuteApplicableStartDate",
                        help_text="Use ISO 8061 (YYYY-MM-DD)",
                        null=True,
                        verbose_name="Statute start date",
                    ),
                ),
                (
                    "statuteapplicableenddate",
                    models.TextField(
                        blank=True,
                        db_column="statuteApplicableEndDate",
                        help_text="Use ISO 8061 (YYYY-MM-DD)",
                        null=True,
                        verbose_name="Statute end date",
                    ),
                ),
                (
                    "statuteenddateopen",
                    models.BooleanField(
                        db_column="statuteApplicableEndDateOpen",
                        default=False,
                        help_text="Indicate end date is open",
                        verbose_name="Open End Date",
                    ),
                ),
                (
                    "rightsstatement",
                    models.ForeignKey(
                        db_column="fkRightsStatement",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="main.rightsstatement",
                    ),
                ),
            ],
            options={
                "verbose_name": "Rights: Statute",
                "db_table": "RightsStatementStatuteInformation",
            },
        ),
        migrations.CreateModel(
            name="TransferMetadataField",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        db_column="pk",
                        default=uuid.uuid4,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "createdtime",
                    models.DateTimeField(
                        auto_now_add=True, db_column="createdTime", null=True
                    ),
                ),
                (
                    "fieldlabel",
                    models.CharField(blank=True, db_column="fieldLabel", max_length=50),
                ),
                ("fieldname", models.CharField(db_column="fieldName", max_length=50)),
                ("fieldtype", models.CharField(db_column="fieldType", max_length=50)),
                ("sortorder", models.IntegerField(db_column="sortOrder", default=0)),
            ],
            options={
                "db_table": "TransferMetadataFields",
            },
        ),
        migrations.CreateModel(
            name="TransferMetadataSet",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        db_column="pk",
                        default=uuid.uuid4,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "createdtime",
                    models.DateTimeField(auto_now_add=True, db_column="createdTime"),
                ),
                ("createdbyuserid", models.IntegerField(db_column="createdByUserID")),
            ],
            options={
                "db_table": "TransferMetadataSets",
            },
        ),
        migrations.CreateModel(
            name="UnitVariable",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        db_column="pk",
                        default=uuid.uuid4,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "unittype",
                    models.CharField(
                        blank=True, db_column="unitType", max_length=50, null=True
                    ),
                ),
                (
                    "unituuid",
                    models.CharField(
                        db_column="unitUUID",
                        help_text="Semantically a foreign key to SIP or Transfer",
                        max_length=36,
                        null=True,
                    ),
                ),
                ("variable", models.TextField(db_column="variable", null=True)),
                (
                    "variablevalue",
                    models.TextField(db_column="variableValue", null=True),
                ),
                (
                    "microservicechainlink",
                    models.UUIDField(
                        blank=True,
                        db_column="microServiceChainLink",
                        default=uuid.uuid4,
                        null=True,
                    ),
                ),
                (
                    "createdtime",
                    models.DateTimeField(auto_now_add=True, db_column="createdTime"),
                ),
                (
                    "updatedtime",
                    models.DateTimeField(auto_now=True, db_column="updatedTime"),
                ),
            ],
            options={
                "db_table": "UnitVariables",
            },
        ),
        migrations.CreateModel(
            name="TransferMetadataFieldValue",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        db_column="pk",
                        default=uuid.uuid4,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "createdtime",
                    models.DateTimeField(auto_now_add=True, db_column="createdTime"),
                ),
                ("fieldvalue", models.TextField(blank=True, db_column="fieldValue")),
                (
                    "field",
                    models.ForeignKey(
                        db_column="fieldUUID",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="main.transfermetadatafield",
                    ),
                ),
                (
                    "set",
                    models.ForeignKey(
                        db_column="setUUID",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="main.transfermetadataset",
                    ),
                ),
            ],
            options={
                "db_table": "TransferMetadataFieldValues",
            },
        ),
        migrations.CreateModel(
            name="Transfer",
            fields=[
                (
                    "uuid",
                    models.CharField(
                        db_column="transferUUID",
                        max_length=36,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("currentlocation", models.TextField(db_column="currentLocation")),
                ("type", models.CharField(db_column="type", max_length=50)),
                ("accessionid", models.TextField(db_column="accessionID")),
                (
                    "sourceofacquisition",
                    models.TextField(blank=True, db_column="sourceOfAcquisition"),
                ),
                (
                    "typeoftransfer",
                    models.TextField(blank=True, db_column="typeOfTransfer"),
                ),
                ("description", models.TextField(blank=True)),
                ("notes", models.TextField(blank=True)),
                ("access_system_id", models.TextField(db_column="access_system_id")),
                ("hidden", models.BooleanField(default=False)),
                ("diruuids", models.BooleanField(db_column="dirUUIDs", default=False)),
                (
                    "transfermetadatasetrow",
                    models.ForeignKey(
                        blank=True,
                        db_column="transferMetadataSetRowUUID",
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="main.transfermetadataset",
                    ),
                ),
            ],
            options={
                "db_table": "Transfers",
            },
        ),
        migrations.CreateModel(
            name="Task",
            fields=[
                (
                    "taskuuid",
                    models.CharField(
                        db_column="taskUUID",
                        max_length=36,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("createdtime", models.DateTimeField(db_column="createdTime")),
                (
                    "fileuuid",
                    models.CharField(
                        blank=True, db_column="fileUUID", max_length=36, null=True
                    ),
                ),
                ("filename", models.TextField(blank=True, db_column="fileName")),
                (
                    "execution",
                    models.CharField(blank=True, db_column="exec", max_length=250),
                ),
                ("arguments", models.CharField(blank=True, max_length=1000)),
                (
                    "starttime",
                    models.DateTimeField(
                        db_column="startTime", default=None, null=True
                    ),
                ),
                (
                    "endtime",
                    models.DateTimeField(db_column="endTime", default=None, null=True),
                ),
                ("client", models.CharField(blank=True, max_length=50)),
                ("stdout", models.TextField(blank=True, db_column="stdOut")),
                ("stderror", models.TextField(blank=True, db_column="stdError")),
                (
                    "exitcode",
                    models.BigIntegerField(blank=True, db_column="exitCode", null=True),
                ),
                (
                    "job",
                    models.ForeignKey(
                        db_column="jobuuid",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="main.job",
                    ),
                ),
            ],
            options={
                "db_table": "Tasks",
            },
        ),
        migrations.CreateModel(
            name="SIP",
            fields=[
                (
                    "uuid",
                    models.CharField(
                        db_column="sipUUID",
                        max_length=36,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "createdtime",
                    models.DateTimeField(auto_now_add=True, db_column="createdTime"),
                ),
                (
                    "currentpath",
                    models.TextField(blank=True, db_column="currentPath", null=True),
                ),
                ("hidden", models.BooleanField(default=False)),
                (
                    "aip_filename",
                    models.TextField(blank=True, db_column="aipFilename", null=True),
                ),
                ("diruuids", models.BooleanField(db_column="dirUUIDs", default=False)),
                ("identifiers", models.ManyToManyField(to="main.identifier")),
            ],
            options={
                "db_table": "SIPs",
            },
        ),
        migrations.CreateModel(
            name="RightsStatementStatuteInformationNote",
            fields=[
                (
                    "id",
                    models.AutoField(db_column="pk", primary_key=True, serialize=False),
                ),
                (
                    "statutenote",
                    models.TextField(
                        db_column="statuteNote", verbose_name="Statute note"
                    ),
                ),
                (
                    "rightsstatementstatute",
                    models.ForeignKey(
                        db_column="fkRightsStatementStatuteInformation",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="main.rightsstatementstatuteinformation",
                    ),
                ),
            ],
            options={
                "verbose_name": "Rights: Statute: Note",
                "db_table": "RightsStatementStatuteInformationNote",
            },
        ),
        migrations.CreateModel(
            name="RightsStatementStatuteDocumentationIdentifier",
            fields=[
                (
                    "id",
                    models.AutoField(
                        db_column="pk",
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "statutedocumentationidentifiertype",
                    models.TextField(
                        db_column="statuteDocumentationIdentifierType",
                        verbose_name="Statute document identification type",
                    ),
                ),
                (
                    "statutedocumentationidentifiervalue",
                    models.TextField(
                        db_column="statuteDocumentationIdentifierValue",
                        verbose_name="Statute document identification value",
                    ),
                ),
                (
                    "statutedocumentationidentifierrole",
                    models.TextField(
                        blank=True,
                        db_column="statuteDocumentationIdentifierRole",
                        null=True,
                        verbose_name="Statute document identification role",
                    ),
                ),
                (
                    "rightsstatementstatute",
                    models.ForeignKey(
                        db_column="fkRightsStatementStatuteInformation",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="main.rightsstatementstatuteinformation",
                    ),
                ),
            ],
            options={
                "verbose_name": "Rights: Statute: Docs ID",
                "db_table": "RightsStatementStatuteDocumentationIdentifier",
            },
        ),
        migrations.CreateModel(
            name="RightsStatementRightsGrantedRestriction",
            fields=[
                (
                    "id",
                    models.AutoField(db_column="pk", primary_key=True, serialize=False),
                ),
                ("restriction", models.TextField(db_column="restriction")),
                (
                    "rightsgranted",
                    models.ForeignKey(
                        db_column="fkRightsStatementRightsGranted",
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="restrictions",
                        to="main.rightsstatementrightsgranted",
                    ),
                ),
            ],
            options={
                "verbose_name": "Rights: Granted: Restriction",
                "db_table": "RightsStatementRightsGrantedRestriction",
            },
        ),
        migrations.CreateModel(
            name="RightsStatementRightsGrantedNote",
            fields=[
                (
                    "id",
                    models.AutoField(
                        db_column="pk",
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "rightsgrantednote",
                    models.TextField(
                        db_column="rightsGrantedNote", verbose_name="Rights note"
                    ),
                ),
                (
                    "rightsgranted",
                    models.ForeignKey(
                        db_column="fkRightsStatementRightsGranted",
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="notes",
                        to="main.rightsstatementrightsgranted",
                    ),
                ),
            ],
            options={
                "verbose_name": "Rights: Granted: Note",
                "db_table": "RightsStatementRightsGrantedNote",
            },
        ),
        migrations.CreateModel(
            name="RightsStatementOtherRightsInformationNote",
            fields=[
                (
                    "id",
                    models.AutoField(db_column="pk", primary_key=True, serialize=False),
                ),
                (
                    "otherrightsnote",
                    models.TextField(
                        db_column="otherRightsNote", verbose_name="Other rights note"
                    ),
                ),
                (
                    "rightsstatementotherrights",
                    models.ForeignKey(
                        db_column="fkRightsStatementOtherRightsInformation",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="main.rightsstatementotherrightsinformation",
                    ),
                ),
            ],
            options={
                "verbose_name": "Rights: Other: Note",
                "db_table": "RightsStatementOtherRightsNote",
            },
        ),
        migrations.CreateModel(
            name="RightsStatementOtherRightsDocumentationIdentifier",
            fields=[
                (
                    "id",
                    models.AutoField(
                        db_column="pk",
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "otherrightsdocumentationidentifiertype",
                    models.TextField(
                        db_column="otherRightsDocumentationIdentifierType",
                        verbose_name="Other rights document identification type",
                    ),
                ),
                (
                    "otherrightsdocumentationidentifiervalue",
                    models.TextField(
                        db_column="otherRightsDocumentationIdentifierValue",
                        verbose_name="Other right document identification value",
                    ),
                ),
                (
                    "otherrightsdocumentationidentifierrole",
                    models.TextField(
                        blank=True,
                        db_column="otherRightsDocumentationIdentifierRole",
                        null=True,
                        verbose_name="Other rights document identification role",
                    ),
                ),
                (
                    "rightsstatementotherrights",
                    models.ForeignKey(
                        db_column="fkRightsStatementOtherRightsInformation",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="main.rightsstatementotherrightsinformation",
                    ),
                ),
            ],
            options={
                "verbose_name": "Rights: Other: Docs ID",
                "db_table": "RightsStatementOtherRightsDocumentationIdentifier",
            },
        ),
        migrations.CreateModel(
            name="RightsStatementLinkingAgentIdentifier",
            fields=[
                (
                    "id",
                    models.AutoField(db_column="pk", primary_key=True, serialize=False),
                ),
                (
                    "linkingagentidentifiertype",
                    models.TextField(
                        blank=True,
                        db_column="linkingAgentIdentifierType",
                        verbose_name="Linking Agent",
                    ),
                ),
                (
                    "linkingagentidentifiervalue",
                    models.TextField(
                        blank=True,
                        db_column="linkingAgentIdentifierValue",
                        verbose_name="Linking Agent Value",
                    ),
                ),
                (
                    "rightsstatement",
                    models.ForeignKey(
                        db_column="fkRightsStatement",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="main.rightsstatement",
                    ),
                ),
            ],
            options={
                "verbose_name": "Rights: Agent",
                "db_table": "RightsStatementLinkingAgentIdentifier",
            },
        ),
        migrations.CreateModel(
            name="RightsStatementLicenseNote",
            fields=[
                (
                    "id",
                    models.AutoField(
                        db_column="pk",
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "licensenote",
                    models.TextField(
                        db_column="licenseNote", verbose_name="License note"
                    ),
                ),
                (
                    "rightsstatementlicense",
                    models.ForeignKey(
                        db_column="fkRightsStatementLicense",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="main.rightsstatementlicense",
                    ),
                ),
            ],
            options={
                "verbose_name": "Rights: License: Note",
                "db_table": "RightsStatementLicenseNote",
            },
        ),
        migrations.CreateModel(
            name="RightsStatementLicenseDocumentationIdentifier",
            fields=[
                (
                    "id",
                    models.AutoField(
                        db_column="pk",
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "licensedocumentationidentifiertype",
                    models.TextField(
                        db_column="licenseDocumentationIdentifierType",
                        verbose_name="License documentation identification type",
                    ),
                ),
                (
                    "licensedocumentationidentifiervalue",
                    models.TextField(
                        db_column="licenseDocumentationIdentifierValue",
                        verbose_name="License documentation identification value",
                    ),
                ),
                (
                    "licensedocumentationidentifierrole",
                    models.TextField(
                        blank=True,
                        db_column="licenseDocumentationIdentifierRole",
                        null=True,
                        verbose_name="License document identification role",
                    ),
                ),
                (
                    "rightsstatementlicense",
                    models.ForeignKey(
                        db_column="fkRightsStatementLicense",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="main.rightsstatementlicense",
                    ),
                ),
            ],
            options={
                "verbose_name": "Rights: License: Docs ID",
                "db_table": "RightsStatementLicenseDocumentationIdentifier",
            },
        ),
        migrations.CreateModel(
            name="RightsStatementCopyrightNote",
            fields=[
                (
                    "id",
                    models.AutoField(
                        db_column="pk",
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "copyrightnote",
                    models.TextField(
                        db_column="copyrightNote", verbose_name="Copyright note"
                    ),
                ),
                (
                    "rightscopyright",
                    models.ForeignKey(
                        db_column="fkRightsStatementCopyrightInformation",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="main.rightsstatementcopyright",
                    ),
                ),
            ],
            options={
                "verbose_name": "Rights: Copyright: Note",
                "db_table": "RightsStatementCopyrightNote",
            },
        ),
        migrations.CreateModel(
            name="RightsStatementCopyrightDocumentationIdentifier",
            fields=[
                (
                    "id",
                    models.AutoField(
                        db_column="pk",
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "copyrightdocumentationidentifiertype",
                    models.TextField(
                        db_column="copyrightDocumentationIdentifierType",
                        verbose_name="Copyright document identification type",
                    ),
                ),
                (
                    "copyrightdocumentationidentifiervalue",
                    models.TextField(
                        db_column="copyrightDocumentationIdentifierValue",
                        verbose_name="Copyright document identification value",
                    ),
                ),
                (
                    "copyrightdocumentationidentifierrole",
                    models.TextField(
                        blank=True,
                        db_column="copyrightDocumentationIdentifierRole",
                        null=True,
                        verbose_name="Copyright document identification role",
                    ),
                ),
                (
                    "rightscopyright",
                    models.ForeignKey(
                        db_column="fkRightsStatementCopyrightInformation",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="main.rightsstatementcopyright",
                    ),
                ),
            ],
            options={
                "verbose_name": "Rights: Copyright: Docs ID",
                "db_table": "RightsStatementCopyrightDocumentationIdentifier",
            },
        ),
        migrations.AddIndex(
            model_name="job",
            index=models.Index(
                fields=["sipuuid", "createdtime", "createdtimedec"],
                name="Jobs_SIPUUID_6cf6c1_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="job",
            index=models.Index(
                fields=["sipuuid", "jobtype", "createdtime", "createdtimedec"],
                name="Jobs_SIPUUID_cf4b11_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="job",
            index=models.Index(
                fields=[
                    "sipuuid",
                    "currentstep",
                    "microservicegroup",
                    "microservicechainlink",
                ],
                name="Jobs_SIPUUID_658a37_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="job",
            index=models.Index(
                fields=["jobtype", "currentstep"], name="Jobs_jobType_4a3346_idx"
            ),
        ),
        migrations.AddField(
            model_name="fpcommandoutput",
            name="file",
            field=models.ForeignKey(
                db_column="fileUUID",
                on_delete=django.db.models.deletion.CASCADE,
                to="main.file",
            ),
        ),
        migrations.AddField(
            model_name="fileid",
            name="file",
            field=models.ForeignKey(
                blank=True,
                db_column="fileUUID",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="main.file",
            ),
        ),
        migrations.AddField(
            model_name="fileformatversion",
            name="file_uuid",
            field=models.ForeignKey(
                db_column="fileUUID",
                on_delete=django.db.models.deletion.CASCADE,
                to="main.file",
            ),
        ),
        migrations.AddField(
            model_name="file",
            name="identifiers",
            field=models.ManyToManyField(to="main.identifier"),
        ),
        migrations.AddField(
            model_name="file",
            name="sip",
            field=models.ForeignKey(
                blank=True,
                db_column="sipUUID",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="main.sip",
            ),
        ),
        migrations.AddField(
            model_name="file",
            name="transfer",
            field=models.ForeignKey(
                blank=True,
                db_column="transferUUID",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="main.transfer",
            ),
        ),
        migrations.AddField(
            model_name="event",
            name="agents",
            field=models.ManyToManyField(to="main.agent"),
        ),
        migrations.AddField(
            model_name="event",
            name="file_uuid",
            field=models.ForeignKey(
                blank=True,
                db_column="fileUUID",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="main.file",
            ),
        ),
        migrations.AddField(
            model_name="dublincore",
            name="metadataappliestotype",
            field=models.ForeignKey(
                db_column="metadataAppliesToType",
                on_delete=django.db.models.deletion.CASCADE,
                to="main.metadataappliestotype",
            ),
        ),
        migrations.AddField(
            model_name="directory",
            name="identifiers",
            field=models.ManyToManyField(to="main.identifier"),
        ),
        migrations.AddField(
            model_name="directory",
            name="sip",
            field=models.ForeignKey(
                blank=True,
                db_column="sipUUID",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="main.sip",
            ),
        ),
        migrations.AddField(
            model_name="directory",
            name="transfer",
            field=models.ForeignKey(
                blank=True,
                db_column="transferUUID",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="main.transfer",
            ),
        ),
        migrations.AddField(
            model_name="derivation",
            name="derived_file",
            field=models.ForeignKey(
                db_column="derivedFileUUID",
                on_delete=django.db.models.deletion.CASCADE,
                related_name="original_file_set",
                to="main.file",
            ),
        ),
        migrations.AddField(
            model_name="derivation",
            name="event",
            field=models.ForeignKey(
                blank=True,
                db_column="relatedEventUUID",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="main.event",
                to_field="event_id",
            ),
        ),
        migrations.AddField(
            model_name="derivation",
            name="source_file",
            field=models.ForeignKey(
                db_column="sourceFileUUID",
                on_delete=django.db.models.deletion.CASCADE,
                related_name="derived_file_set",
                to="main.file",
            ),
        ),
        migrations.AddIndex(
            model_name="file",
            index=models.Index(
                fields=["sip", "filegrpuse"], name="Files_sipUUID_a0d05d_idx"
            ),
        ),
    ]



# This file is part of Archivematica.
#
# Copyright 2010-2013 Artefactual Systems Inc. <http://artefactual.com>
#
# Archivematica is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Archivematica is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Archivematica.  If not, see <http://www.gnu.org/licenses/>.
import os
import re
import shutil

from unidecode import unidecode

from a3m.archivematicaFunctions import strToUnicode


VERSION = "1.10." + "$Id: 3e6f862d213fabb26d88080c74c2d912331dad17 $".split(" ")[1]

# Letters, digits and a few punctuation characters
ALLOWED_CHARS = re.compile(r"[^a-zA-Z0-9\-_.\(\)]")
REPLACEMENT_CHAR = "_"


def change_name(basename):
    if basename == "":
        raise ValueError("change_name recieved an empty filename.")
    unicode_basename = strToUnicode(basename)
    unicode_name = unidecode(unicode_basename)
    # We can't return  an empty string here because it will become the new filename.
    # However, in some cases unidecode just strips out all chars (e.g.
    # unidecode(u"🚀") == ""), so if that happens, we to replace the invalid chars with
    # REPLACEMENT_CHAR. This will result in a filename of one or more underscores,
    # which isn't great, but allows processing to continue.
    if unicode_name == "":
        unicode_name = unicode_basename

    return ALLOWED_CHARS.sub(REPLACEMENT_CHAR, unicode_name)


def change_path(path):
    basename = os.path.basename(path)
    changed_name = change_name(basename)

    if basename == changed_name:
        return path

    dirname = os.path.dirname(path)

    n = 1
    file_title, file_extension = os.path.splitext(changed_name)
    changed_name = os.path.join(dirname, file_title + file_extension)

    while os.path.exists(changed_name):
        changed_name = os.path.join(
            dirname, file_title + REPLACEMENT_CHAR + str(n) + file_extension
        )
        n += 1
    shutil.move(path, changed_name)

    return changed_name


def change_tree(start_path, old_start_path):
    """
    Recursive generator to change all filesystem entries under the start
    path given.

    Yields a tuple of (old_path, changed_path, is_dir, was_changed) once
    for each file or dir within the start_path, everything contained in each
    dir.
    """
    start_path = os.path.abspath(start_path)

    for dir_entry in os.scandir(start_path):
        is_dir = dir_entry.is_dir()  # cache is_dir before rename

        changed_name = change_path(dir_entry.path)
        changed_path = os.path.join(start_path, changed_name)
        old_path = os.path.join(old_start_path, dir_entry.name)

        was_changed = changed_path != old_path
        yield old_path, changed_path, is_dir, was_changed

        if is_dir:
            yield from change_tree(changed_path, old_path)



# This file is part of Archivematica.
#
# Copyright 2010-2013 Artefactual Systems Inc. <http://artefactual.com>
#
# Archivematica is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Archivematica is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Archivematica.  If not, see <http://www.gnu.org/licenses/>.
import os

from django.db import transaction

from a3m.fileOperations import rename
from a3m.main.models import Transfer


def updateDB(dst, transferUUID):
    Transfer.objects.filter(uuid=transferUUID).update(currentlocation=dst)


def moveSIP(job, src, dst, transferUUID, sharedDirectoryPath):
    # os.rename(src, dst)
    if src.endswith("/"):
        src = src[:-1]

    dest = dst.replace(sharedDirectoryPath, "%sharedPath%", 1)
    if dest.endswith("/"):
        dest = os.path.join(dest, os.path.basename(src))
    if dest.endswith("/."):
        dest = os.path.join(dest[:-1], os.path.basename(src))

    if os.path.isdir(src):
        dest += "/"
    updateDB(dest, transferUUID)

    return rename(src, dst, printfn=job.pyprint, should_exit=False)


def call(jobs):
    with transaction.atomic():
        for job in jobs:
            with job.JobContext():
                src = job.args[1]
                dst = job.args[2]
                transferUUID = job.args[3]
                sharedDirectoryPath = job.args[4]
                job.set_status(
                    moveSIP(job, src, dst, transferUUID, sharedDirectoryPath)
                )



# This file is part of Archivematica.
#
# Copyright 2010-2013 Artefactual Systems Inc. <http://artefactual.com>
#
# Archivematica is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Archivematica is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Archivematica.  If not, see <http://www.gnu.org/licenses/>.
import os
from optparse import OptionParser

from django.db import transaction

from a3m.main.models import File


def something(job, SIPDirectory, serviceDirectory, objectsDirectory, SIPUUID, date):
    # exitCode = 435
    exitCode = 0
    job.pyprint(SIPDirectory)
    # For every file, & directory Try to find the matching file & directory in the objects directory
    for path, dirs, files in os.walk(serviceDirectory):
        for file in files:
            servicePreExtension = "_me"
            originalPreExtension = "_m"
            file1Full = os.path.join(path, file).replace(
                SIPDirectory, "%SIPDirectory%", 1
            )  # service

            a = file.rfind(servicePreExtension + ".")
            if a != -1:
                file2Full = os.path.join(
                    path, file[:a] + originalPreExtension + "."
                ).replace(
                    SIPDirectory + "objects/service/", "%SIPDirectory%objects/", 1
                )  # service
            else:
                a = file.rfind(".")
                if a != -1:  # if a period is found
                    a += 1  # include the period
                file2Full = os.path.join(path, file[:a]).replace(
                    SIPDirectory + "objects/service/", "%SIPDirectory%objects/", 1
                )  # service

            f = File.objects.get(
                currentlocation=file1Full, removedtime__isnull=True, sip_id=SIPUUID
            )
            f.filegrpuse = "service"

            grp_file = File.objects.get(
                currentlocation__startswith=file2Full,
                removedtime__isnull=True,
                sip_id=SIPUUID,
            )
            f.filegrpuuid = grp_file.uuid
            f.save()

    return exitCode


def call(jobs):
    parser = OptionParser()
    # '--SIPDirectory "%SIPDirectory%" --serviceDirectory "objects/service/" --objectsDirectory "objects/" --SIPUUID "%SIPUUID%" --date "%date%"' );
    parser.add_option(
        "-s", "--SIPDirectory", action="store", dest="SIPDirectory", default=""
    )
    parser.add_option("-u", "--SIPUUID", action="store", dest="SIPUUID", default="")
    parser.add_option(
        "-a", "--serviceDirectory", action="store", dest="serviceDirectory", default=""
    )
    parser.add_option(
        "-o", "--objectsDirectory", action="store", dest="objectsDirectory", default=""
    )
    parser.add_option("-t", "--date", action="store", dest="date", default="")

    with transaction.atomic():
        for job in jobs:
            with job.JobContext():
                (opts, args) = parser.parse_args(job.args[1:])

                SIPDirectory = opts.SIPDirectory
                serviceDirectory = os.path.join(SIPDirectory, opts.serviceDirectory)
                objectsDirectory = os.path.join(SIPDirectory, opts.objectsDirectory)
                SIPUUID = opts.SIPUUID
                date = opts.date

                if not os.path.isdir(serviceDirectory):
                    job.pyprint("no service directory in this sip")
                    job.set_status(0)
                    continue

                exitCode = something(
                    job, SIPDirectory, serviceDirectory, objectsDirectory, SIPUUID, date
                )
                job.set_status(exitCode)



# This file is part of Archivematica.
#
# Copyright 2010-2013 Artefactual Systems Inc. <http://artefactual.com>
#
# Archivematica is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Archivematica is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Archivematica.  If not, see <http://www.gnu.org/licenses/>.
import argparse
import logging
import os
import uuid

from django.db import transaction

from a3m.fileOperations import get_size_and_checksum
from a3m.fileOperations import updateSizeAndChecksum
from a3m.main.models import File


logger = logging.getLogger(__name__)

SIP_REPLACEMENT_PATH_STRING = r"%SIPDirectory%"
TRANSFER_REPLACEMENT_PATH_STRING = r"%transferDirectory%"


def _filter_queryset_by_subdir(queryset, replacement_path_string, filter_subdir):
    """Filter queryset by filter_subdir."""
    filter_path = "".join([replacement_path_string, filter_subdir])
    return queryset.filter(currentlocation__startswith=filter_path)


def get_transfer_file_queryset(transfer_uuid, filter_subdir):
    """Return Queryset of files in this transfer."""
    files = File.objects.filter(transfer=transfer_uuid)
    if filter_subdir:
        files = _filter_queryset_by_subdir(
            files, TRANSFER_REPLACEMENT_PATH_STRING, filter_subdir
        )
    return files


def get_sip_file_queryset(sip_uuid, filter_subdir):
    """Return Queryset of files in this SIP."""
    files = File.objects.filter(sip=sip_uuid)
    if filter_subdir:
        files = _filter_queryset_by_subdir(
            files, SIP_REPLACEMENT_PATH_STRING, filter_subdir
        )
    return files


def get_size_and_checksum_for_file(file_, sip_directory, transfer_uuid):
    """Get size and checksum for a file."""
    if transfer_uuid:
        file_path = file_.currentlocation.replace(
            TRANSFER_REPLACEMENT_PATH_STRING, sip_directory
        )
    else:
        file_path = file_.currentlocation.replace(
            SIP_REPLACEMENT_PATH_STRING, sip_directory
        )

    if not os.path.exists(file_path):
        return {}

    fileSize, checksum, checksumType = get_size_and_checksum(file_path)

    return {
        "filePath": file_path,
        "fileSize": fileSize,
        "checksum": checksum,
        "checksumType": checksumType,
    }


def call(jobs):
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-s", "--sipDirectory", action="store", dest="sip_directory", default=""
    )
    parser.add_argument("-S", "--sipUUID", type=uuid.UUID, dest="sip_uuid")
    parser.add_argument("-T", "--transferUUID", type=uuid.UUID, dest="transfer_uuid")
    parser.add_argument("-d", "--date", action="store", dest="date", default="")
    parser.add_argument(
        "--filterSubdir", action="store", dest="filter_subdir", default=None
    )
    parser.add_argument(
        "-u",
        "--eventIdentifierUUID",
        type=uuid.UUID,
        dest="event_uuid",
    )

    state = []

    for job in jobs:
        with job.JobContext(logger=logger):
            args = parser.parse_args(job.args[1:])

            TRANSFER_SIP_UUIDS = [args.sip_uuid, args.transfer_uuid]
            if all(TRANSFER_SIP_UUIDS) or not any(TRANSFER_SIP_UUIDS):
                job.print_error("SIP exclusive-or Transfer UUID must be defined")
                job.set_status(2)
                continue

            files = get_transfer_file_queryset(args.transfer_uuid, args.filter_subdir)
            if args.sip_uuid:
                files = get_sip_file_queryset(args.sip_uuid, args.filter_subdir)

            for file_ in files:
                if not file_:
                    continue
                file_info = get_size_and_checksum_for_file(
                    file_, args.sip_directory, args.transfer_uuid
                )
                if file_info:
                    state.append((file_.uuid, file_info, args))

            job.set_status(0)

    with transaction.atomic():
        for file_uuid, file_info, args in state:
            file_path = file_info.pop("filePath")
            updateSizeAndChecksum(
                file_uuid, file_path, args.date, args.event_uuid, **file_info
            )



"""
Job chain logic.

The `JobChain` class determines which job should be run next, based on
the workflow and the exit code of the previous job. It also stores context
provided by previous jobs in the same chain (e.g. choices for a decision).

This module knows about all `Job` subclasses, and will instantiate the correct
one by looking at the workflow.
"""
import logging
from collections.abc import Iterator

from a3m.server.jobs.base import Job
from a3m.server.jobs.client import ClientScriptJob
from a3m.server.jobs.client import DirectoryClientScriptJob
from a3m.server.jobs.client import FilesClientScriptJob
from a3m.server.jobs.decisions import NextLinkDecisionJob


logger = logging.getLogger(__name__)


def get_job_class_for_link(link):
    """Return the appropriate `Job` class to be used for the workflow link.

    Keys off of the `@manager` config option, which referenced the previous
    linkTaskManager implementation.
    """
    manager_name = link.config["@manager"]
    if manager_name == "linkTaskManagerDirectories":
        job_class = DirectoryClientScriptJob
    elif manager_name == "linkTaskManagerFiles":
        job_class = FilesClientScriptJob
    elif manager_name == "linkTaskManagerChoice":
        job_class = NextLinkDecisionJob
    else:
        raise ValueError(f"Unknown manager type {manager_name}")

    return job_class


class JobChain:
    """
    Creates jobs as necessary based on the workflow chain and package given.

    Intended to be used as an iterator that returns jobs (e.g.
    `job = next(JobChain(...))`).

    Job chains are used for passing information between jobs, via:
        * context, a dict of replacement variables for tasks
        * next_link, a workflow link that can be set to redirect the job chain
    """

    def __init__(self, package, workflow, starting_link):
        """Create an instance of a chain, based on the workflow chain given."""
        self.package = package
        self.workflow = workflow

        self.initial_link = starting_link
        self.current_link = None
        self.next_link = self.initial_link

        self.current_job = None
        # TODO: package context hits the db, make that clearer
        self.context = self.package.context.copy()

        logger.debug(
            "Creating JobChain for package %s (initial link %s)",
            package.uuid,
            self.initial_link,
        )

    def __iter__(self) -> Iterator[Job]:
        return self

    def __next__(self):
        if self.next_link:
            next_link = self.next_link
            self.next_link = None
        else:
            try:
                next_link = self.current_link.get_next_link(self.current_job.exit_code)
            except KeyError:
                next_link = None

        # End of chain.
        if not next_link:
            self.current_link = None
            self.current_job = None
            self.chain_completed()
            raise StopIteration

        # Ensure we have a Link instance instead of its identifier.
        if isinstance(next_link, str):
            next_link = self.workflow.get_link(next_link)

        self.current_link = next_link
        job_class = get_job_class_for_link(self.current_link)
        self.current_job = job_class(self, self.current_link, self.package)
        return self.current_job

    def job_completed(self):
        logger.debug(
            "%s %s done with exit code %s",
            self.current_job.__class__.__name__,
            self.current_job.uuid,
            self.current_job.exit_code,
        )
        if isinstance(self.current_job, ClientScriptJob):
            self.current_job.update_status_from_exit_code()
        else:
            self.current_job.mark_complete()

    def chain_completed(self):
        """Log chain completion"""
        logger.debug("Done with chain for package %s", self.package.uuid)



"""
Start and run MCPServer.

`main` goes through the following steps:
1. A `ThreadPoolExecutor` is initialized with a configurable number of threads
(default ncpus).
2. A signal listener is setup to handle shutdown on SIGINT/SIGTERM events.
3. The default workflow is loaded (from workflow.json).
4. The configured SHARED_DIRECTORY is populated with the expected directory
structure, and default processing configs added.
5. Any in progress Job and Task entries in the database are marked as errors,
as they are presumed to have been the result of a shutdown while processing.
6. If Prometheus metrics are enabled, an thread is started to serve metrics for
scraping.
7. A `PackageQueue` (see the `queues` module) is initialized.
8. A configured number (default 4) of RPCServer (see the `rpc_server` module)
threads are started to handle gearman "RPC" requests from the dashboard.
9. A watched directory thread is started to observe changes in any of the
watched dirs as set in the workflow.
10. The `PackageQueue.work` processing loop is started on the main thread.
"""
import concurrent.futures
import enum
import logging
import threading

import grpc
from django.conf import settings as premis_settings
from grpc_reflection.v1alpha import reflection

from a3m import __version__
from a3m.api.transferservice import v1beta1 as transfer_service_api
from a3m.main import models
from a3m.server import metrics
from a3m.server import shared_dirs
from a3m.server.db import migrate
from a3m.server.jobs import Job
from a3m.server.queues import PackageQueue
from a3m.server.tasks import Task
from a3m.server.tasks.backends import get_task_backend
from a3m.server.tasks.backends import TaskBackend
from a3m.server.transfer_service import TransferService
from a3m.server.workflow import load_default_workflow
from a3m.server.workflow import Workflow

logger = logging.getLogger(__name__)


@enum.unique
class ServerStage(enum.Enum):
    STOPPED = "stopped"
    STARTED = "started"
    GRACE = "grace"


class Server:
    """a3m server.

    It runs the gRPC API server and the workflow engine, using independent pools
    of threads. It accepts a :class:`a3m.server.workflow.Workflow` which can be
    customized as needed.
    """

    def __init__(
        self,
        bind_address: str,
        server_credentials: grpc.ServerCredentials | None,
        workflow: Workflow,
        max_concurrent_packages: int,
        batch_size: int,
        queue_executor: concurrent.futures.ThreadPoolExecutor,
        grpc_executor: concurrent.futures.ThreadPoolExecutor,
        debug: bool = False,
    ):
        self.stage = ServerStage.STOPPED
        self.lock = threading.RLock()
        self.termination_event = threading.Event()

        # A3M-TODO: inject task backend via PackageQueue.
        TaskBackend.TASK_BATCH_SIZE = batch_size

        self.workflow = workflow
        self.queue_executor = queue_executor
        self.queue_shutdown_event = threading.Event()
        self.queue = PackageQueue(
            self.queue_executor,
            self.queue_shutdown_event,
            max_concurrent_packages=max_concurrent_packages,
            debug=debug,
        )
        self.grpc_executor = grpc_executor
        self.grpc_server = grpc.server(grpc_executor)
        self.grpc_port = self.grpc_server.add_insecure_port(bind_address)

        self._mount_services()

    def _mount_services(self):
        transfer_service = TransferService(
            self.workflow, self.queue, self.queue_executor
        )
        transfer_service_api.service_pb2_grpc.add_TransferServiceServicer_to_server(
            transfer_service, self.grpc_server
        )

        services = tuple(
            service.full_name
            for service in transfer_service_api.service_pb2.DESCRIPTOR.services_by_name.values()
        ) + (reflection.SERVICE_NAME,)
        reflection.enable_server_reflection(services, self.grpc_server)

    def start(self):
        """Starts this Server.

        This method may only be called once. (i.e. it is not idempotent).
        """
        with self.lock:
            if self.stage is not ServerStage.STOPPED:
                raise ValueError("Cannot start already-started server!")

            self.stage = ServerStage.STARTED

            threading.Thread(target=self.queue.work).start()
            threading.Thread(target=self.grpc_server.start).start()

    def wait_for_termination(self, timeout=None):
        """Blocks current thread until the server stops."""
        while not self.termination_event.is_set():
            self.termination_event.wait(timeout=None)
        return False

    def stop(self, grace=None):
        """Stops this Server."""
        with self.lock:
            shutdown_event = threading.Event()

            if self.stage is ServerStage.STOPPED:
                shutdown_event.set()
                return shutdown_event

            self.stage = ServerStage.GRACE

            def _stop():
                logger.info("Shutting down...")

                self.grpc_server.stop(grace)
                self.queue_shutdown_event.set()
                self.queue.wait_for_termination()
                get_task_backend().shutdown(wait=False)

                shutdown_event.set()
                self.termination_event.set()
                self.stage = ServerStage.STOPPED

            threading.Thread(target=_stop).start()

            shutdown_event.wait()
            return shutdown_event


def create_server(
    bind_address,
    server_credentials,
    max_concurrent_packages,
    batch_size,
    queue_workers,
    grpc_workers,
    debug=False,
):
    """Create a3m server ready to use.

    It bootstraps some bits locally needed, like the database, the local
    processing directory or the pool of threads. It wraps
    :class:`a3m.server.runner.Server`.
    """
    workflow = load_default_workflow()

    shared_dirs.create()

    migrate()
    update_agents()

    Job.cleanup_old_db_entries()
    Task.cleanup_old_db_entries()

    metrics.init_labels(workflow)
    metrics.start_prometheus_server()

    return Server(
        bind_address,
        server_credentials,
        workflow,
        max_concurrent_packages,
        batch_size,
        concurrent.futures.ThreadPoolExecutor(max_workers=queue_workers),
        concurrent.futures.ThreadPoolExecutor(max_workers=grpc_workers),
        debug,
    )


def update_agents():
    """Create or update software and organization agents."""

    models.Agent.objects.update_or_create(
        agenttype="software",
        name="a3m",
        identifiertype="preservation system",
        defaults={
            "identifiervalue": f"a3m version={__version__}",
        },
    )

    if premis_settings.ORG_ID and premis_settings.ORG_NAME:
        models.Agent.objects.update_or_create(
            agenttype="organization",
            identifiertype="repository code",
            defaults={
                "identifiervalue": premis_settings.ORG_ID,
                "name": premis_settings.ORG_NAME,
            },
        )



# This file is part of Archivematica.
#
# Copyright 2010-2013 Artefactual Systems Inc. <http://artefactual.com>
#
# Archivematica is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Archivematica is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Archivematica.  If not, see <http://www.gnu.org/licenses/>.
import os
import sys


def call(jobs):
    for job in jobs:
        with job.JobContext():
            target = job.args[1]
            if not os.path.isdir(target):
                job.pyprint("Directory doesn't exist: ", target, file=sys.stderr)
                os.mkdir(target)
            if os.listdir(target) == []:
                job.pyprint("Directory is empty: ", target, file=sys.stderr)
                fileName = os.path.join(target, "submissionDocumentation.log")
                f = open(fileName, "a")
                f.write("No submission documentation added")
                f.close()
                os.chmod(fileName, 0o600)



from argparse import ArgumentParser

from lxml import etree

from a3m.client import metrics
from a3m.main.models import Transfer


def fetch_set(sip_uuid):
    transfer = Transfer.objects.get(uuid=sip_uuid)
    return transfer.transfermetadatasetrow


def fetch_fields_and_values(sip_uuid):
    metadata_set = fetch_set(sip_uuid)
    if metadata_set is None:
        return []

    results = metadata_set.transfermetadatafieldvalue_set.exclude(
        fieldvalue=""
    ).values_list("field__fieldname", "fieldvalue")

    return results


def build_element(label, value, root):
    element = etree.SubElement(root, label)
    element.text = value
    return element


def call(jobs):
    parser = ArgumentParser(
        description="Create a generic XML document from transfer metadata"
    )
    parser.add_argument("-S", "--sipUUID", action="store", dest="sip_uuid")
    parser.add_argument("-x", "--xmlFile", action="store", dest="xml_file")

    for job in jobs:
        with job.JobContext():
            opts = parser.parse_args(job.args[1:])

            root = etree.Element("transfer_metadata")

            values = fetch_fields_and_values(opts.sip_uuid)
            elements = [build_element(label, value, root) for (label, value) in values]

            # If there is no transfer metadata, skip writing the XML
            if elements:
                tree = etree.ElementTree(root)
                tree.write(
                    opts.xml_file,
                    pretty_print=True,
                    xml_declaration=True,
                    encoding="utf-8",
                )

                job.pyprint(etree.tostring(tree))

            # This is an odd point to mark the transfer as "completed", but it's the
            # last step in the "Complete Transfer" microservice group before the folder
            # move, so it seems like the best option we have for now.
            metrics.transfer_completed(opts.sip_uuid)



#
# This file is part of Archivematica.
#
# Copyright 2010-2012 Artefactual Systems Inc. <http://artefactual.com>
#
# Archivematica is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Archivematica is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Archivematica.    If not, see <http://www.gnu.org/licenses/>.
import collections
import csv
import sys
import traceback

from a3m import archivematicaFunctions


def parseMetadata(job, SIPPath, state):
    """
    Parse all metadata.csv files in SIPPath.

    Looking for metadata.csvs in metadata/ and
    objects/metadata/transfers/<transfer name>/metadata/

    See parseMetadataCSV for details on parsing.

    :param SIPPath: Path to the SIP
    :return: {<filename>: OrderedDict(key: [values]) }
    """
    all_metadata = {}
    metadata_csvs = archivematicaFunctions.find_metadata_files(SIPPath, "metadata.csv")

    for metadataCSVFilePath in metadata_csvs:
        try:
            csv_metadata = parseMetadataCSV(job, metadataCSVFilePath)
        except Exception:
            job.pyprint("error parsing: ", metadataCSVFilePath, file=sys.stderr)
            job.print_error(traceback.format_exc())
            state.error_accumulator.error_count += 1
            continue
        # Provide warning if this file already has differing metadata
        # Not using all_metadata.update(csv_metadata) because of that
        for entry, values in csv_metadata.items():
            if entry in all_metadata and all_metadata[entry] != values:
                job.pyprint(
                    "Metadata for",
                    entry,
                    "being updated. Old:",
                    all_metadata[entry],
                    "New:",
                    values,
                    file=sys.stderr,
                )
            existing = all_metadata.get(entry, collections.OrderedDict())
            existing.update(values)
            all_metadata[entry] = existing

    return all_metadata


def parseMetadataCSV(job, metadataCSVFilePath):
    """
    Parses the metadata.csv into a dict with entries for each file.

    Each file's entry is an OrderedDict containing the column header and a list of values for each column.

    Example CSV:
    Filename,dc.title,dc.type,dc.type,Other metadata
    objects/foo.jpg,Foo,Photograph,Still Image,Taken on a sunny day
    objects/bar/,Bar,Photograph,Still Image,All taken on a rainy day

    Produces:
    {
        'objects/foo.jpg': OrderedDict(dc.title=[Foo], dc.type=[Photograph, Still Image], Other metadata=[Taken on a sunny day])
        'objects/bar': OrderedDict(dc.title=[Bar], dc.date=[Photograph, Still Image], Other metadata=[All taken on a rainy day])
    }

    :param metadataCSVFilePath: Path to the metadata CSV to parse
    :return: {<filename>: OrderedDict(<metadata name>: [<metadata value>]) }
    """
    metadata = {}
    # use universal newline mode to support unusual newlines, like \r
    with open(metadataCSVFilePath) as f:
        reader = csv.reader(f)
        # Parse first row as header
        header = next(reader)
        # Strip filename column, strip whitespace from header values
        header = [h.strip() for h in header[1:]]
        # Parse data
        for row in reader:
            if not row:
                continue
            entry_name = row[0]
            if entry_name.endswith("/"):
                entry_name = entry_name[:-1]
            # Strip file/dir name from values
            row = row[1:]
            values = archivematicaFunctions.OrderedListsDict(list(zip(header, row)))
            if entry_name in metadata and metadata[entry_name] != values:
                job.pyprint(
                    "Metadata for",
                    entry_name,
                    "being overwritten. Old:",
                    metadata[entry_name],
                    "New:",
                    values,
                    file=sys.stderr,
                )
            metadata[entry_name] = values

    return collections.OrderedDict(metadata)  # Return a normal OrderedDict



import argparse
import logging
import uuid

import pygfried
from django.db import transaction

from a3m.databaseFunctions import getUTCDate
from a3m.databaseFunctions import insertIntoEvents
from a3m.fpr.registry import FormatVersion
from a3m.fpr.registry import FPR
from a3m.main.models import File
from a3m.main.models import FileID


logger = logging.getLogger(__name__)

TOOL_DESCRIPTION = "pygfried/siegfried"
TOOL_VERSION = pygfried.version()


def write_file_format_version(file_obj: File, format_version_id: str):
    (file_format_version, created) = file_obj.fileformatversion_set.get_or_create(
        defaults={"format_version_id": format_version_id}
    )
    if not created:  # Update the version if it wasn't created new
        file_format_version.format_version_id = format_version_id
        file_format_version.save()


def write_identification_event(
    file_id: str, puid: str | None = None, success: bool = True
):
    event_detail_text = 'program="{}"; version="{}"'.format(
        TOOL_DESCRIPTION, TOOL_VERSION
    )
    if success:
        event_outcome_text = "Positive"
    else:
        event_outcome_text = "Not identified"

    if not puid or puid == "UNKNOWN":
        puid = "No Matching Format"

    date = getUTCDate()

    insertIntoEvents(
        fileUUID=file_id,
        eventIdentifierUUID=str(uuid.uuid4()),
        eventType="format identification",
        eventDateTime=date,
        eventDetail=event_detail_text,
        eventOutcome=event_outcome_text,
        eventOutcomeDetailNote=puid,
    )


def write_file_id(file_id: str, format_version: FormatVersion):
    """
    Write the identified format to the DB.
    """
    FileID.objects.create(
        file_id=file_id,
        format_name=format_version.format.description,
        format_version=format_version.version or "",
        format_registry_name="PRONOM",
        format_registry_key=format_version.pronom_id,
    )


def main(file_path: str, file_id: str, disable_reidentify: bool) -> int:
    # If reidentification is disabled and a format identification event exists for this file, exit
    file_obj = File.objects.get(uuid=file_id)
    if (
        disable_reidentify
        and file_obj.event_set.filter(event_type="format identification").exists()
    ):
        logger.debug(
            "This file has already been identified, and re-identification is disabled. Skipping."
        )
        return 0

    puid: str | None = None
    try:
        puid = pygfried.identify(file_path)
    except Exception as err:
        logger.error("Error running pygfried: %s", err)
        return 255

    if not puid or puid == "UNKNOWN":
        write_identification_event(file_id, success=False)
        return 255

    format_version = FPR.get_format_version_by_puid(puid)
    if not format_version:
        write_identification_event(file_id, success=False)
        return 255

    write_file_format_version(file_obj, str(format_version.id))
    write_identification_event(file_id, puid=puid)
    write_file_id(file_id, format_version)

    return 0


def call(jobs):
    parser = argparse.ArgumentParser(description="Identify file formats.")
    parser.add_argument("file_path", type=str, help="%relativeLocation%")
    parser.add_argument("file_uuid", type=str, help="%fileUUID%")
    parser.add_argument(
        "--disable-reidentify",
        action="store_true",
        help="Disable identification if it has already happened for this file.",
    )

    with transaction.atomic():
        for job in jobs:
            with job.JobContext():
                args = parser.parse_args(job.args[1:])
                job.set_status(
                    main(
                        args.file_path,
                        args.file_uuid,
                        args.disable_reidentify,
                    )
                )



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from a3m.api.transferservice.v1beta1 import (
    request_response_pb2 as a3m_dot_api_dot_transferservice_dot_v1beta1_dot_request__response__pb2,
)


class TransferServiceStub:
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Submit = channel.unary_unary(
            "/a3m.api.transferservice.v1beta1.TransferService/Submit",
            request_serializer=a3m_dot_api_dot_transferservice_dot_v1beta1_dot_request__response__pb2.SubmitRequest.SerializeToString,
            response_deserializer=a3m_dot_api_dot_transferservice_dot_v1beta1_dot_request__response__pb2.SubmitResponse.FromString,
        )
        self.Read = channel.unary_unary(
            "/a3m.api.transferservice.v1beta1.TransferService/Read",
            request_serializer=a3m_dot_api_dot_transferservice_dot_v1beta1_dot_request__response__pb2.ReadRequest.SerializeToString,
            response_deserializer=a3m_dot_api_dot_transferservice_dot_v1beta1_dot_request__response__pb2.ReadResponse.FromString,
        )
        self.ListTasks = channel.unary_unary(
            "/a3m.api.transferservice.v1beta1.TransferService/ListTasks",
            request_serializer=a3m_dot_api_dot_transferservice_dot_v1beta1_dot_request__response__pb2.ListTasksRequest.SerializeToString,
            response_deserializer=a3m_dot_api_dot_transferservice_dot_v1beta1_dot_request__response__pb2.ListTasksResponse.FromString,
        )
        self.Empty = channel.unary_unary(
            "/a3m.api.transferservice.v1beta1.TransferService/Empty",
            request_serializer=a3m_dot_api_dot_transferservice_dot_v1beta1_dot_request__response__pb2.EmptyRequest.SerializeToString,
            response_deserializer=a3m_dot_api_dot_transferservice_dot_v1beta1_dot_request__response__pb2.EmptyResponse.FromString,
        )


class TransferServiceServicer:
    """Missing associated documentation comment in .proto file."""

    def Submit(self, request, context):
        """Submits a new transfer."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def Read(self, request, context):
        """Reads the status of a given transfer."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ListTasks(self, request, context):
        """Lists all tasks in a given transfer."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def Empty(self, request, context):
        """Delete all contents from a3m's shared folders. Should only be called once processing is complete."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_TransferServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "Submit": grpc.unary_unary_rpc_method_handler(
            servicer.Submit,
            request_deserializer=a3m_dot_api_dot_transferservice_dot_v1beta1_dot_request__response__pb2.SubmitRequest.FromString,
            response_serializer=a3m_dot_api_dot_transferservice_dot_v1beta1_dot_request__response__pb2.SubmitResponse.SerializeToString,
        ),
        "Read": grpc.unary_unary_rpc_method_handler(
            servicer.Read,
            request_deserializer=a3m_dot_api_dot_transferservice_dot_v1beta1_dot_request__response__pb2.ReadRequest.FromString,
            response_serializer=a3m_dot_api_dot_transferservice_dot_v1beta1_dot_request__response__pb2.ReadResponse.SerializeToString,
        ),
        "ListTasks": grpc.unary_unary_rpc_method_handler(
            servicer.ListTasks,
            request_deserializer=a3m_dot_api_dot_transferservice_dot_v1beta1_dot_request__response__pb2.ListTasksRequest.FromString,
            response_serializer=a3m_dot_api_dot_transferservice_dot_v1beta1_dot_request__response__pb2.ListTasksResponse.SerializeToString,
        ),
        "Empty": grpc.unary_unary_rpc_method_handler(
            servicer.Empty,
            request_deserializer=a3m_dot_api_dot_transferservice_dot_v1beta1_dot_request__response__pb2.EmptyRequest.FromString,
            response_serializer=a3m_dot_api_dot_transferservice_dot_v1beta1_dot_request__response__pb2.EmptyResponse.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        "a3m.api.transferservice.v1beta1.TransferService", rpc_method_handlers
    )
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class TransferService:
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Submit(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/a3m.api.transferservice.v1beta1.TransferService/Submit",
            a3m_dot_api_dot_transferservice_dot_v1beta1_dot_request__response__pb2.SubmitRequest.SerializeToString,
            a3m_dot_api_dot_transferservice_dot_v1beta1_dot_request__response__pb2.SubmitResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def Read(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/a3m.api.transferservice.v1beta1.TransferService/Read",
            a3m_dot_api_dot_transferservice_dot_v1beta1_dot_request__response__pb2.ReadRequest.SerializeToString,
            a3m_dot_api_dot_transferservice_dot_v1beta1_dot_request__response__pb2.ReadResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ListTasks(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/a3m.api.transferservice.v1beta1.TransferService/ListTasks",
            a3m_dot_api_dot_transferservice_dot_v1beta1_dot_request__response__pb2.ListTasksRequest.SerializeToString,
            a3m_dot_api_dot_transferservice_dot_v1beta1_dot_request__response__pb2.ListTasksResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def Empty(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/a3m.api.transferservice.v1beta1.TransferService/Empty",
            a3m_dot_api_dot_transferservice_dot_v1beta1_dot_request__response__pb2.EmptyRequest.SerializeToString,
            a3m_dot_api_dot_transferservice_dot_v1beta1_dot_request__response__pb2.EmptyResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )



import os
from uuid import uuid4

from django.db import transaction
from django.utils import timezone

from a3m import databaseFunctions
from a3m import fileOperations
from a3m.dicts import ReplacementDict
from a3m.dicts import setup_dicts
from a3m.executeOrRunSubProcess import executeOrRun
from a3m.fpr.registry import FPR
from a3m.fpr.registry import Rule
from a3m.fpr.registry import RulePurpose
from a3m.main.models import Derivation
from a3m.main.models import File


def insert_transcription_event(status, file_uuid, rule, relative_location):
    outcome = "transcribed" if status == 0 else "not transcribed"

    tool = rule.command.tool
    event_detail = 'program={}; version={}; command="{}"'.format(
        tool.description, tool.version, rule.command.command.replace('"', r"\"")
    )

    event_uuid = str(uuid4())

    databaseFunctions.insertIntoEvents(
        fileUUID=file_uuid,
        eventIdentifierUUID=event_uuid,
        eventType="transcription",
        eventDetail=event_detail,
        eventOutcome=outcome,
        eventOutcomeDetailNote=relative_location,
    )

    return event_uuid


def insert_file_into_database(
    task_uuid, file_uuid, sip_uuid, event_uuid, rule, output_path, relative_path
):
    transcription_uuid = str(uuid4())
    today = timezone.now()
    fileOperations.addFileToSIP(
        relative_path,
        transcription_uuid,
        sip_uuid,
        task_uuid,
        today,
        sourceType="creation",
        use="text/ocr",
    )

    fileOperations.updateSizeAndChecksum(
        transcription_uuid, output_path, today, str(uuid4())
    )

    databaseFunctions.insertIntoDerivations(
        sourceFileUUID=file_uuid,
        derivedFileUUID=transcription_uuid,
        relatedEventUUID=event_uuid,
    )


def fetch_rules_for(file_obj) -> list[Rule]:
    return FPR.get_file_rules(file_obj, purpose=RulePurpose.TRANSCRIPTION)


def fetch_rules_for_derivatives(file_obj) -> tuple[File | None, list[Rule]]:
    derivs = Derivation.objects.filter(source_file=file_obj)
    for deriv in derivs:
        derived_file = deriv.derived_file
        rules = fetch_rules_for(derived_file)
        if rules:
            return (derived_file, rules)

    return None, []


def main(job, task_uuid, file_uuid):
    setup_dicts()

    succeeded = True

    file_obj = File.objects.get(uuid=file_uuid)

    # Normally we don't transcribe derivatives (access copies, preservation copies);
    # however, some useful transcription tools can't handle some formats that
    # are common as the primary copies. For example, tesseract can't handle JPEG2000.
    # If there are no rules for the primary format passed in, try to look at each
    # derivative until a transcribable derivative is found.
    #
    # Skip derivatives to avoid double-scanning them; only look at them as a fallback.
    if file_obj.filegrpuse != "original":
        job.print_error(f"{file_uuid} is not an original; not transcribing")
        return 0

    rules = fetch_rules_for(file_obj)
    if not rules:
        file_obj, rules = fetch_rules_for_derivatives(file_obj)

    if not rules:
        job.print_error(
            "No rules found for file {} and its derivatives; not transcribing".format(
                file_uuid
            )
        )
        return 0
    else:
        if file_obj.filegrpuse == "original":
            noun = "original"
        else:
            noun = file_obj.filegrpuse + " derivative"
        job.print_error(f"Transcribing {noun} {file_obj.uuid}")

    rd = ReplacementDict.frommodel(file_=file_obj, type_="file")

    for rule in rules:
        script = rule.command.command
        if rule.command.script_type.value in ("bashScript", "command"):
            (script,) = rd.replace(script)
            args = []
        else:
            args = rd.to_gnu_options

        exitstatus, stdout, stderr = executeOrRun(
            rule.command.script_type, script, arguments=args, capture_output=True
        )
        job.write_output(stdout)
        job.write_error(stderr)
        if exitstatus != 0:
            succeeded = False

        output_path = rd.replace(rule.command.output_location)[0]
        relative_path = output_path.replace(rd["%SIPDirectory%"], "%SIPDirectory%")
        event = insert_transcription_event(exitstatus, file_uuid, rule, relative_path)

        if os.path.isfile(output_path):
            insert_file_into_database(
                task_uuid,
                file_uuid,
                rd["%SIPUUID%"],
                event,
                rule,
                output_path,
                relative_path,
            )

    return 0 if succeeded else 1


def call(jobs):
    with transaction.atomic():
        for job in jobs:
            with job.JobContext():
                task_uuid = job.args[1]
                file_uuid = job.args[2]

                job.set_status(main(job, task_uuid, file_uuid))
